#include <linux/fs.h>
#include <linux/vmalloc.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/mm.h>
#include <linux/string.h>
#include <linux/kernel.h>
#include <linux/proc_fs.h>
#include <linux/ctype.h>
#include <linux/pagemap.h>
#include <linux/usb.h>


#include "spca_isoc.h"
#include "spcadecoder.h"
#include "zc3xx.h"


static unsigned long spca5xx_gettimes(void);
static void spca50x_isoc_irq(struct urb *urb, struct pt_regs *regs);
static int spca50x_move_data(struct usb_spca50x *spca50x, struct urb *urb);
static void outpict_do_tasklet(unsigned long ptr);
static struct spca50x_frame *spca50x_next_frame(struct usb_spca50x
						*spca50x, unsigned char *cdata);
/**********************************************************************
* spca50x_stop_isoc
* Function stops the USB ISO pipe by stopping the chip ISO machine
* and stopping USB transfer
***********************************************************************/
void spca5xx_kill_isoc(struct usb_spca50x *spca50x)
{
	int n;

	if (!spca50x)
		return;

	PDEBUG(3, "*** killing capture ***");
	spca50x->streaming = 0;

	/* Unschedule all of the iso td's */
	for (n = SPCA50X_NUMSBUF - 1; n >= 0; n--) {
		if (spca50x->sbuf[n].urb) {
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,8)
			usb_kill_urb(spca50x->sbuf[n].urb);
#else
			usb_unlink_urb(spca50x->sbuf[n].urb);
#endif
			usb_free_urb(spca50x->sbuf[n].urb);
			spca50x->sbuf[n].urb = NULL;
		}
	}

	PDEBUG(3, "*** Isoc killed ***");
}

void spca50x_stop_isoc(struct usb_spca50x *spca50x)
{

	if (!spca50x->streaming || !spca50x->dev)
		return;

	PDEBUG(3, "*** Stopping capture ***");
	//spcaCameraStop(spca50x);
	zc3xx_stop(spca50x);
	spca5xx_kill_isoc(spca50x);
	spca50x_set_packet_size(spca50x, 0);

	PDEBUG(3, "*** Capture stopped ***");
}

int spca50x_set_packet_size(struct usb_spca50x *spca50x, int size)
{
	int alt;
	/**********************************************************************/
	/******** Try to find real Packet size from usb struct ****************/
	struct usb_device *dev = spca50x->dev;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
	struct usb_interface_descriptor *interface = NULL;

	struct usb_config_descriptor *config = dev->actconfig;
#else
	struct usb_host_interface *interface = NULL;
	struct usb_interface *intf;
#endif
	int mysize = 0;
	int ep = 0;

	/**********************************************************************/

	if (size == 0)
		alt = SPCA50X_ALT_SIZE_0;
	else if (size == 128)
		alt = SPCA50X_ALT_SIZE_128;
	else if (size == 256)
		alt = SPCA50X_ALT_SIZE_256;
	else if (size == 384)
		alt = SPCA50X_ALT_SIZE_384;
	else if (size == 512)
		alt = SPCA50X_ALT_SIZE_512;
	else if (size == 640)
		alt = SPCA50X_ALT_SIZE_640;
	else if (size == 768)
		alt = SPCA50X_ALT_SIZE_768;
	else if (size == 896)
		alt = SPCA50X_ALT_SIZE_896;
	else if (size == 1023)
		alt = SPCA50X_ALT_SIZE_1023;
	else {
		/* if an unrecognised size, default to the minimum */
		PDEBUG(5,
				"Set packet size: invalid size (%d), defaulting to %d",
				size, SPCA50X_ALT_SIZE_128);
		alt = SPCA50X_ALT_SIZE_128;
	}

	PDEBUG(5, "iface alt size: %d %d %d", spca50x->iface, alt, size);
	if (usb_set_interface(spca50x->dev, spca50x->iface, alt) < 0) {
		err("Set packet size: set interface error");
		return -EBUSY;
	}
	/* check endpoint */
	if (spca50x->bridge == BRIDGE_PAC207) {
		ep = PAC207_ENDPOINT_ADDRESS - 1;
	} else {
		ep = SPCA50X_ENDPOINT_ADDRESS - 1;
	}

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,3)
	intf = usb_ifnum_to_if(dev, spca50x->iface);
	if (intf) {
		interface = usb_altnum_to_altsetting(intf, alt);
	} else {
		PDEBUG(0, "intf not found");
		return -ENXIO;
	}
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,11)
	mysize = le16_to_cpu(interface->endpoint[ep].desc.wMaxPacketSize);
#else
	mysize = (interface->endpoint[ep].desc.wMaxPacketSize);
#endif
#else
	interface = &config->interface[spca50x->iface].altsetting[alt];
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,11)
	mysize = le16_to_cpu(interface->endpoint[ep].wMaxPacketSize);
#else
	mysize = (interface->endpoint[ep].wMaxPacketSize);
#endif
#endif
	
	spca50x->packet_size = mysize & 0x03ff; /* < 1023 */
	spca50x->alt = alt;
	PDEBUG(1, "[check] set real packet size: %d, alt=%d", mysize, alt);
	return 0;
}
/* Tasklet function to decode */
static void outpict_do_tasklet(unsigned long ptr)
{
	int err;
	struct spca50x_frame *taskletframe = (struct spca50x_frame *)ptr;

	taskletframe->scanlength = taskletframe->highwater - taskletframe->data;
	PDEBUG(4,
	       "Tasklet ask spcadecoder hdrwidth %d hdrheight %d method %d ",
	       taskletframe->hdrwidth, taskletframe->hdrheight,
	       taskletframe->method);

	if ((err = spca50x_outpicture(taskletframe)) < 0) {
		PDEBUG(2, "frame decoder failed (%d)", err);
		taskletframe->grabstate = FRAME_ERROR;
	} else {
		taskletframe->grabstate = FRAME_DONE;
		PDEBUG(4, "Decode framestate return %d",
		       taskletframe->grabstate);
	}

	if (waitqueue_active(&taskletframe->wq))
		wake_up_interruptible(&taskletframe->wq);
}

int spca50x_init_isoc(struct usb_spca50x *spca50x)
{
	struct urb *urb;
	int fx, err, n;
	//int intpipe;
	PDEBUG(3, "*** Initializing capture ***");
/* reset iso context */
	spca50x->compress = 0;
	spca50x->curframe = 0;
	spca50x->cursbuf = 0;
	spca50x->frame[0].seq = -1;
	spca50x->lastFrameRead = -1;

	PDEBUG(3, ">>>> spca50x_init_isoc");
	spca50x_set_packet_size(spca50x, spca50x->pipe_size);
	PDEBUG(2, "setpacketsize %d", spca50x->pipe_size);

	for (n = 0; n < SPCA50X_NUMSBUF; n++) {
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,5,0)
		urb = usb_alloc_urb(FRAMES_PER_DESC, GFP_KERNEL);
#else
		urb = usb_alloc_urb(FRAMES_PER_DESC);
#endif
		if (!urb) {
			err("init isoc: usb_alloc_urb ret. NULL");
			return -ENOMEM;
		}
		spca50x->sbuf[n].urb = urb;
		urb->dev = spca50x->dev;
		urb->context = spca50x;
		if (spca50x->bridge == BRIDGE_PAC207) {
			urb->pipe =
			    usb_rcvisocpipe(spca50x->dev,
					    PAC207_ENDPOINT_ADDRESS);
		} else {
			urb->pipe =
			    usb_rcvisocpipe(spca50x->dev,
					    SPCA50X_ENDPOINT_ADDRESS);
		}
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)
		urb->transfer_flags = URB_ISO_ASAP;
		urb->interval = 1;
#else
		urb->transfer_flags = USB_ISO_ASAP;
#endif
		urb->transfer_buffer = spca50x->sbuf[n].data;
		urb->complete = spca50x_isoc_irq;
		urb->number_of_packets = FRAMES_PER_DESC;
		urb->transfer_buffer_length =
		    spca50x->packet_size * FRAMES_PER_DESC;
		for (fx = 0; fx < FRAMES_PER_DESC; fx++) {
			urb->iso_frame_desc[fx].offset =
			    spca50x->packet_size * fx;
			urb->iso_frame_desc[fx].length = spca50x->packet_size;
		}
	}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,41)
	spca50x->sbuf[SPCA50X_NUMSBUF - 1].urb->next = spca50x->sbuf[0].urb;
	for (n = 0; n < SPCA50X_NUMSBUF - 1; n++)
		spca50x->sbuf[n].urb->next = spca50x->sbuf[n + 1].urb;
#endif
	//spcaCameraStart(spca50x);
	zc3xx_start(spca50x);
	PDEBUG(0, "init isoc int %d altsetting %d", spca50x->iface,
	       spca50x->alt);
	for (n = 0; n < SPCA50X_NUMSBUF; n++) {
		spca50x->sbuf[n].urb->dev = spca50x->dev;
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,5,41)
		err = usb_submit_urb(spca50x->sbuf[n].urb, GFP_KERNEL);
#else
		err = usb_submit_urb(spca50x->sbuf[n].urb);
#endif
		if (err) {
			err("init isoc: usb_submit_urb(%d) ret %d", n, err);
			return err;
		}
	}

	for (n = 0; n < SPCA50X_NUMFRAMES; n++) {
		spca50x->frame[n].grabstate = FRAME_UNUSED;
		spca50x->frame[n].scanstate = STATE_SCANNING;
	}

	spca50x->streaming = 1;

	return 0;

}

/**********************************************************************
* spca50x_isoc_irq
* Function processes the finish of the USB transfer by calling 
* spca50x_move_data function to move data from USB buffer to internal
* driver structures 
***********************************************************************/
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,5,0)
static void spca50x_isoc_irq(struct urb *urb, struct pt_regs *regs)
{
	int i;
#else
static void spca50x_isoc_irq(struct urb *urb)
{
#endif
	int len;
	struct usb_spca50x *spca50x;

	if (!urb->context) {
		PDEBUG(4, "no context");
		return;
	}

	spca50x = (struct usb_spca50x *)urb->context;

	if (!spca50x->dev) {
		PDEBUG(4, "no device ");
		return;
	}
	if (!spca50x->user) {
		PDEBUG(4, "device not open");
		return;
	}
	if (!spca50x->streaming) {
		/* Always get some of these after close but before packet engine stops */
		PDEBUG(4, "hmmm... not streaming, but got interrupt");
		return;
	}
	if (!spca50x->present) {
		/*  */
		PDEBUG(4, "device disconnected ..., but got interrupt !!");
		return;
	}
	/* Copy the data received into our scratch buffer */
	if (spca50x->curframe >= 0) {
		len = spca50x_move_data(spca50x, urb);
	} else if (waitqueue_active(&spca50x->wq)) {
		wake_up_interruptible(&spca50x->wq);
	}

	/* Move to the next sbuf */
	spca50x->cursbuf = (spca50x->cursbuf + 1) % SPCA50X_NUMSBUF;

	urb->dev = spca50x->dev;
	urb->status = 0;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 5, 4)
	if ((i = usb_submit_urb(urb, GFP_ATOMIC)) != 0)
		err("usb_submit_urb() ret %d", i);
#else
/* If we use urb->next usb_submit_urb() is not need in 2.4.x */
	//if ((i = usb_submit_urb(urb)) != 0)
	//err("usb_submit_urb() ret %d", i);
#endif

	return;
}

/* ******************************************************************
* spca50x_move_data
* Function serves for moving data from USB transfer buffers
* to internal driver frame buffers.
******************************************************************* */
static int spca50x_move_data(struct usb_spca50x *spca50x, struct urb *urb)
{
	unsigned char *cdata;	//Pointer to buffer where we do store next packet
	unsigned char *pData;	//Pointer to buffer where we do store next packet
	int i;
	struct spca50x_frame *frame;	//Pointer to frame data
	int iPix;		//Offset of pixel data in the ISO packet

	unsigned long ms_times_now;
	unsigned long ms_times_before;
	int totlen = 0;
	int tv8532 = 0;

	for (i = 0; i < urb->number_of_packets; i++) {
		int datalength = urb->iso_frame_desc[i].actual_length;
		int st = urb->iso_frame_desc[i].status;
		int sequenceNumber;
		//int seqframe = 0;
		//int sof;
		//int j, p;
		/* PDEBUG(5,"Packet data [%d,%d,%d]", datalength, st,
		   urb->iso_frame_desc[i].offset); */

		urb->iso_frame_desc[i].actual_length = 0;
		urb->iso_frame_desc[i].status = 0;

		cdata = ((unsigned char *)urb->transfer_buffer) +
		    urb->iso_frame_desc[i].offset;
		/* Check for zero length block or no selected frame buffer */
		if (!datalength || spca50x->curframe == -1) {
			tv8532 = 0;
			continue;
		}

		PDEBUG(5, "Packet data [%d,%d,%d] Status: %d", datalength, st,
		       urb->iso_frame_desc[i].offset, st);

		if (st)
			PDEBUG(2, "data error: [%d] len=%d, status=%d", i,
			       datalength, st);

		frame = &spca50x->frame[spca50x->curframe];
		totlen = frame->totlength;

		/* read the sequence number */
		if (spca50x->bridge == BRIDGE_ZC3XX) {
			if (frame->last_packet == -1) {
				/*initialize a new frame */
				sequenceNumber = 0;
			} else {
				sequenceNumber = frame->last_packet;
			}
		} else {
			sequenceNumber = cdata[SPCA50X_OFFSET_SEQUENCE];
			PDEBUG(4, "Packet start  %x %x %x %x %x", cdata[0],
			       cdata[1], cdata[2], cdata[3], cdata[4]);
		}
		/* check frame start */
		switch (spca50x->bridge) {
		case BRIDGE_ZC3XX:
			{
				iPix = 0;
				if (cdata[0] == 0xFF && cdata[1] == 0xD8) {
					sequenceNumber = 0;
					spca50x->header_len = 2;	//18 remove 0xff 0xd8;
					PDEBUG(5,
					       "Zc301 header packet found datalength %d totlength %d!!",
					       datalength, totlen);
				} else {
					sequenceNumber++;
				}

			}
			break;
		default:
			{
				iPix = 1;
				/* check if this is a drop packet */
				if (sequenceNumber == SPCA50X_SEQUENCE_DROP) {
					PDEBUG(3,
					       "Dropped packet (expected seq 0x%02x)",
					       frame->last_packet + 1);
					continue;
				}
			}
			break;
		}

		PDEBUG(4, "spca50x: Packet seqnum = 0x%02x.  curframe=%2d",
		       sequenceNumber, spca50x->curframe);
		pData = cdata;

		/* Can we find a frame start */
		if (sequenceNumber == 0) {
			totlen = 0;
			iPix = spca50x->header_len;
			PDEBUG(4, "spca50x: Found Frame Start!, framenum = %d",
			       spca50x->curframe);
			// Start of frame is implicit end of previous frame
			// Check for a previous frame and finish it off if one exists
			if (frame->scanstate == STATE_LINES) {

				if (frame->format != VIDEO_PALETTE_RAW_JPEG) {
					ms_times_now = spca5xx_gettimes();
					/* overflow ? */
					if (ms_times_now < spca50x->last_times)
						spca50x->last_times = 0;
					spin_lock(&spca50x->v4l_lock);
					ms_times_before =
					    spca50x->last_times +
					    spca50x->dtimes;
					spin_unlock(&spca50x->v4l_lock);
					if (ms_times_now >= ms_times_before) {
						PDEBUG(4,
						       "Decode frame last %d",
						       (int)(ms_times_now -
							     spca50x->
							     last_times));
						spca50x->last_times =
						    ms_times_now;
						/* Decode the frame */
						tasklet_init(&spca50x->
							     spca5xx_tasklet,
							     outpict_do_tasklet,
							     (unsigned long)
							     frame);
						tasklet_schedule(&spca50x->
								 spca5xx_tasklet);
					} else
						frame->grabstate = FRAME_ERROR;
				} else {
					/* RAW DATA stream */
					frame->grabstate = FRAME_DONE;
					if (waitqueue_active(&frame->wq))
						wake_up_interruptible(&frame->
								      wq);
				}
				// If someone decided to wait for ANY frame - wake him up 
				if (waitqueue_active(&spca50x->wq))
					wake_up_interruptible(&spca50x->wq);
				frame = spca50x_next_frame(spca50x, cdata);
			} else
				frame->scanstate = STATE_LINES;
		}

		/* Are we in a frame? */
		if (frame == NULL || frame->scanstate != STATE_LINES)
			continue;
		if (sequenceNumber != frame->last_packet + 1 &&
		    frame->last_packet != -1) {
			/* Note, may get one of these for the first packet after opening */
			PDEBUG(2, "Out of order packet, last = %d, this = %d",
			       frame->last_packet, sequenceNumber);

		}
		frame->last_packet = sequenceNumber;

		/* This is the real conversion of the raw camera data to BGR for V4L */
		if (!(spca50x->bridge == BRIDGE_TV8532 ||
		      spca50x->bridge == BRIDGE_ETOMS))
			datalength -= iPix;	// correct length for packet header

		pData = cdata + iPix;	// Skip packet header (1 or 10 bytes)

		// Consume data
		PDEBUG(5, "Processing packet seq  %d,length %d,totlength %d",
		       frame->last_packet, datalength, frame->totlength);
		/* this copy consume input data from the isoc stream */
		if ((datalength > 0) && (datalength <= 0x3ff)) {
			memcpy(frame->highwater, pData, datalength);
			frame->highwater += datalength;
			totlen += datalength;
		}

		frame->totlength = totlen;
#ifdef SPCA50X_ENABLE_EXP_BRIGHTNESS
		/* autoadjust is set to 1 to so now autobrightness will be
		   calculated frome this frame. autoadjust will be set to 0 when
		   autobrightness has been corrected if needed. */
		autoadjust = 1;
#endif				/* SPCA50X_ENABLE_EXP_BRIGHTNESS */

	}
	return totlen;
}

static struct spca50x_frame *spca50x_next_frame(struct usb_spca50x
						*spca50x, unsigned char *cdata)
{
	int iFrameNext;
	struct spca50x_frame *frame = NULL;
	PDEBUG(4, "Frame %d State %d", spca50x->curframe,
	       spca50x->frame[spca50x->curframe].grabstate);
	if (spca50x->frame[spca50x->curframe].grabstate == FRAME_ERROR) {
		PDEBUG(2, "Frame %d errdrop", spca50x->curframe);
		frame = &spca50x->frame[spca50x->curframe];
		goto dropframe;
	}
	/* Cycle through the frame buffer looking for a free frame to overwrite */
	iFrameNext = (spca50x->curframe + 1) % SPCA50X_NUMFRAMES;
	while (frame == NULL && iFrameNext != (spca50x->curframe)) {
		if (spca50x->frame[iFrameNext].grabstate == FRAME_READY ||
		    spca50x->frame[iFrameNext].grabstate == FRAME_UNUSED ||
		    spca50x->frame[iFrameNext].grabstate == FRAME_ERROR) {
			spca50x->curframe = iFrameNext;
			frame = &spca50x->frame[iFrameNext];
			break;
		} else {
			iFrameNext = (iFrameNext + 1) % SPCA50X_NUMFRAMES;
		}
	}

	if (frame == NULL) {
		//PDEBUG(3, "Can't find a free frame to grab into...using next. "
		 //      "This is caused by the application not reading fast enough.");
		spca50x->curframe = (spca50x->curframe + 1) % SPCA50X_NUMFRAMES;
		frame = &spca50x->frame[spca50x->curframe];
	}
      dropframe:
	frame->grabstate = FRAME_GRABBING;

	/* Record the frame sequence number the camera has told us */
	if (cdata) {
		switch (spca50x->bridge) {
		case BRIDGE_ZC3XX:
			frame->seq = 00;
			break;
		}
	}
	if (spca50x->pictsetting.change) {
		memcpy(&frame->pictsetting, &spca50x->pictsetting,
		       sizeof(struct pictparam));
		/* reset flag change */
		spca50x->pictsetting.change = 0;
		PDEBUG(2, "Picture setting change Pass to decoding   ");
	}
	/* Reset some per-frame variables */
	frame->highwater = frame->data;
	frame->scanstate = STATE_LINES;
	frame->scanlength = 0;
	frame->last_packet = -1;
	frame->totlength = 0;
	spca50x->packet = 0;
	return frame;
}

/*********************************************************************
	time Helper function
**********************************************************************/
static unsigned long spca5xx_gettimes(void)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 0)
	struct timeval tv;
	do_gettimeofday(&tv);
	return (tv.tv_sec * 1000 + tv.tv_usec / 1000);
#else
	u64 times_now;
	times_now = get_jiffies_64();
	return jiffies_to_msecs(times_now);
#endif
}
