#include <linux/fs.h>
#include <linux/vmalloc.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/mm.h>
#include <linux/string.h>
#include <linux/kernel.h>
#include <linux/proc_fs.h>
#include <linux/ctype.h>
#include <linux/pagemap.h>
#include <linux/usb.h>


#include "spca_mem.h"

/**********************************************************************
 *
 * Memory management
 *
 * This is a shameless copy from the USB-cpia driver (linux kernel
 * version 2.3.29 or so, I have no idea what this code actually does ;).
 * Actually it seems to be a copy of a shameless copy of the bttv-driver.
 * Or that is a copy of a shameless copy of ... (To the powers: is there
 * no generic kernel-function to do this sort of stuff?)
 *
 * Yes, it was a shameless copy from the bttv-driver. IIRC, Alan says
 * there will be one, but apparentely not yet -jerdfelt
 *
 * So I copied it again for the ov511 driver -claudio
 * And again for the spca50x driver -jcrisp
 **********************************************************************/
void *rvmalloc(unsigned long size)
{
	void *mem;
	unsigned long adr;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 4, 23)
	unsigned long page;
#endif
	size = PAGE_ALIGN(size);
	mem = vmalloc_32(size);
	if (!mem)
		return NULL;

	memset(mem, 0, size);	/* Clear the ram out, no junk to the user */
	adr = (unsigned long)mem;
	while ((long)size > 0) {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 5, 68)
		SetPageReserved(vmalloc_to_page((void *)adr));
#else
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 4, 23)
		mem_map_reserve(vmalloc_to_page((void *)adr));
#else
		page = kvirt_to_pa(adr);
		mem_map_reserve(VIRT_TO_PAGE(__va(page)));
#endif
#endif
		adr += PAGE_SIZE;
		size -= PAGE_SIZE;
	}

	return mem;
}

void rvfree(void *mem, unsigned long size)
{
	unsigned long adr;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 4, 23)
	unsigned long page;
#endif
	if (!mem)
		return;

	adr = (unsigned long)mem;
	while ((long)size > 0) {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 5, 68)
		ClearPageReserved(vmalloc_to_page((void *)adr));
#else
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 4, 23)
		mem_map_unreserve(vmalloc_to_page((void *)adr));
#else
		page = kvirt_to_pa(adr);
		mem_map_unreserve(VIRT_TO_PAGE(__va(page)));
#endif
#endif
		adr += PAGE_SIZE;
		size -= PAGE_SIZE;
	}
	vfree(mem);
}

/****************************************************************************
 *
 * Buffer management
 *
 ***************************************************************************/
int spca50x_alloc(struct usb_spca50x *spca50x)
{
	int i;

	PDEBUG(4, "entered");
	down(&spca50x->buf_lock);
	spca50x->tmpBuffer = rvmalloc(MAX_FRAME_SIZE);
	if (spca50x->buf_state == BUF_ALLOCATED)
		goto out;

	spca50x->fbuf = rvmalloc(SPCA50X_NUMFRAMES * MAX_DATA_SIZE);
	if (!spca50x->fbuf)
		goto error;

	for (i = 0; i < SPCA50X_NUMFRAMES; i++) {
		spca50x->frame[i].tmpbuffer = spca50x->tmpBuffer;
		spca50x->frame[i].decoder = &spca50x->maindecode;	//connect each frame to the main data decoding 
		spca50x->frame[i].grabstate = FRAME_UNUSED;
		spca50x->frame[i].scanstate = STATE_SCANNING;
		spca50x->frame[i].data = spca50x->fbuf + i * MAX_DATA_SIZE;
		spca50x->frame[i].highwater = spca50x->frame[i].data;
		memset(&spca50x->frame[i].pictsetting, 0,
		       sizeof(struct pictparam));
		PDEBUG(4, "frame[%d] @ %p", i, spca50x->frame[i].data);
	}

	for (i = 0; i < SPCA50X_NUMSBUF; i++) {
		spca50x->sbuf[i].data = kmalloc(FRAMES_PER_DESC *
						MAX_FRAME_SIZE_PER_DESC,
						GFP_KERNEL);
		if (!spca50x->sbuf[i].data)
			goto error;
		PDEBUG(4, "sbuf[%d] @ %p", i, spca50x->sbuf[i].data);
	}
	spca50x->buf_state = BUF_ALLOCATED;
      out:
	up(&spca50x->buf_lock);
	PDEBUG(4, "leaving");
	return 0;
      error:
	/* FIXME: IMHO, it's better to move error deallocation code here. */

	for (i = 0; i < SPCA50X_NUMSBUF; i++) {

		if (spca50x->sbuf[i].data) {
			kfree(spca50x->sbuf[i].data);
			spca50x->sbuf[i].data = NULL;
		}
	}
	if (spca50x->fbuf) {
		rvfree(spca50x->fbuf, SPCA50X_NUMFRAMES * MAX_DATA_SIZE);
		spca50x->fbuf = NULL;
	}
	if (spca50x->tmpBuffer) {
		rvfree(spca50x->tmpBuffer, MAX_FRAME_SIZE);
		spca50x->tmpBuffer = NULL;
	}

	spca50x->buf_state = BUF_NOT_ALLOCATED;
	up(&spca50x->buf_lock);
	PDEBUG(1, "errored");
	return -ENOMEM;
}

void spca5xx_dealloc(struct usb_spca50x *spca50x)
{
	int i;
	PDEBUG(2, "entered dealloc");
	down(&spca50x->buf_lock);
	if (spca50x->fbuf) {
		rvfree(spca50x->fbuf, SPCA50X_NUMFRAMES * MAX_DATA_SIZE);
		spca50x->fbuf = NULL;
		for (i = 0; i < SPCA50X_NUMFRAMES; i++)
			spca50x->frame[i].data = NULL;
	}

	if (spca50x->tmpBuffer) {
		rvfree(spca50x->tmpBuffer, MAX_FRAME_SIZE);
		spca50x->tmpBuffer = NULL;
	}

	for (i = 0; i < SPCA50X_NUMSBUF; i++) {
		if (spca50x->sbuf[i].data) {
			kfree(spca50x->sbuf[i].data);
			spca50x->sbuf[i].data = NULL;
		}
	}

	PDEBUG(2, "buffer memory deallocated");
	spca50x->buf_state = BUF_NOT_ALLOCATED;
	up(&spca50x->buf_lock);
	PDEBUG(2, "leaving dealloc");
}


