#ifndef __ZC3XX_MEM_H__
#define __ZC3XX_MEM_H__
#include "spca5xx.h"

#define MAX_FRAME_SIZE (640 * 480 * 4)
#define MAX_DATA_SIZE (MAX_FRAME_SIZE + sizeof(struct timeval))


int spca50x_alloc(struct usb_spca50x *spca50x);
void spca5xx_dealloc(struct usb_spca50x *spca50x);
void *rvmalloc(unsigned long size);
void rvfree(void *mem, unsigned long size);


/* Given PGD from the address space's page table, return the kernel
 * virtual mapping of the physical memory mapped at ADR.
 */
#ifndef RH9_REMAP
static inline unsigned long uvirt_to_kva(pgd_t * pgd, unsigned long adr)
{
	unsigned long ret = 0UL;
	pmd_t *pmd;
	pte_t *ptep, pte;

	if (!pgd_none(*pgd)) {
#if PUD_SHIFT
		pud_t *pud = pud_offset(pgd, adr);
		if (!pud_none(*pud)) {
			pmd = pmd_offset(pud, adr);
#else
		pmd = pmd_offset(pgd, adr);
#endif
		if (!pmd_none(*pmd)) {
			ptep = pte_offset_kernel(pmd, adr);
			pte = *ptep;
			if (pte_present(pte)) {
				ret =
				    (unsigned long)page_address(pte_page(pte));
				ret |= (adr & (PAGE_SIZE - 1));
			}
#if PUD_SHIFT
		}
#endif
		}
	}

	return ret;
}
#endif				/* RH9_REMAP */
/* Here we want the physical address of the memory.
 * This is used when initializing the contents of the
 * area and marking the pages as reserved.
 */
#ifdef RH9_REMAP
static inline unsigned long kvirt_to_pa(unsigned long adr)
{
	unsigned long kva, ret;

	kva = (unsigned long)page_address(vmalloc_to_page((void *)adr));
	kva |= adr & (PAGE_SIZE - 1);	/* restore the offset */
	ret = __pa(kva);
	return ret;
}

#else				/* RH9_REMAP */
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,5,0)
static inline unsigned long kvirt_to_pa(unsigned long adr)
{
	unsigned long kva, ret;

	kva = (unsigned long)page_address(vmalloc_to_page((void *)adr));
	kva |= adr & (PAGE_SIZE - 1);
	ret = __pa(kva);
	return ret;
}
#else
static inline unsigned long kvirt_to_pa(unsigned long adr)
{
	unsigned long va, kva, ret;

	va = VMALLOC_VMADDR(adr);
	kva = uvirt_to_kva(pgd_offset_k(va), va);
	ret = __pa(kva);
	return ret;
}
#endif
#endif				/* RH9_REMAP */

#endif /* __ZC3XX_MEM_H__ */
