#include "spca_mode.h"
#include "spca_isoc.h"
#include "zc3xx.h"


int spca5xx_get_depth(struct usb_spca50x *spca50x, int palette)
{
	switch (palette) {
	case VIDEO_PALETTE_RGB565:
		return 16;
	case VIDEO_PALETTE_RGB24:
		return 24;
	case VIDEO_PALETTE_YUYV:
		return 16;
	case VIDEO_PALETTE_YUV420P:
		return 12;	/* strange need 12 this break the read method for this planar mode (6*8/4) */
	case VIDEO_PALETTE_RGB32:
		return 32;
	case VIDEO_PALETTE_RAW_JPEG:
		return 24;	/* raw jpeg. what should we return ?? */
	case VIDEO_PALETTE_JPEG:
		if (spca50x->cameratype == JPEG ||
		    spca50x->cameratype == JPGH ||
		    spca50x->cameratype == JPGC ||
		    spca50x->cameratype == JPGS
		    || spca50x->cameratype == JPGM) {
			return 8;
		} else
			return 0;
	default:
		return 0;	/* Invalid format */
	}
}

/* ⺻ ̹   */
int spca5xx_getDefaultMode(struct usb_spca50x *spca50x)
{
	/* set_zc3xxVGA()::CIF::320*240 MODE */
	spca50x->width = 320;
	spca50x->height = 240;
	spca50x->method = 0;
	spca50x->pipe_size = 1023;
	spca50x->mode = 1;
	return 0;
}

/*    */
int spca5xx_setMode(struct usb_spca50x *spca50x, int width, int height, int format)
{
	int i, j;
	int formatIn;

	PDEBUG(2, "Set mode asked w %d h %d p %d", width, height, format);

	if ((formatIn = v4l_to_spca5xx(format)) < 0)
		return -EINVAL;
	
	spca50x->width=width;
	spca50x->height=height;
	spca50x->format=format;
	spca50x->hdrwidth=width;
	spca50x->hdrheight=height;
	PDEBUG(2, "Found code %d method %d", spca50x->mode, spca50x->method);
	PDEBUG(2, "Soft Win width height %d x %d", spca50x->width,
	       spca50x->height);
	PDEBUG(2, "Hard Win width height %d x %d", spca50x->hdrwidth,
	       spca50x->hdrheight);

	for (i = 0; i < SPCA50X_NUMFRAMES; i++) {
		spca50x->frame[i].method = spca50x->method;
		spca50x->frame[i].cameratype = spca50x->cameratype;
		spca50x->frame[i].cropx1 = 0;
		spca50x->frame[i].cropx2 = 0;
		spca50x->frame[i].cropy1 = 0;
		spca50x->frame[i].cropy2 = 0;
		spca50x->frame[i].x = 0;
		spca50x->frame[i].y = 0;
		spca50x->frame[i].hdrwidth = spca50x->hdrwidth;
		spca50x->frame[i].hdrheight = spca50x->hdrheight;
		spca50x->frame[i].width = spca50x->width;
		spca50x->frame[i].height = spca50x->height;
		spca50x->frame[i].format = spca50x->format;
		spca50x->frame[i].scanlength =
		    spca50x->width * spca50x->height * 3 / 2;
		// ?? assumes 4:2:0 data
	}

	return 0;
}

/* reset image mode */
int spca5xx_restartMode(struct usb_spca50x *spca50x, int width, int height,
		    int format)
{
	int was_streaming;
	int r;
	
	if (spca50x->width == width && spca50x->height == height)
		return 0;
	PDEBUG(1, "Mode changing to %d,%d", width, height);

	was_streaming = spca50x->streaming;

	if (was_streaming) {
		spca50x_stop_isoc(spca50x);
	}

	r = spca5xx_setMode(spca50x, width, height, format);
	if (r < 0)
		goto out;
	if (was_streaming) {
		r = spca50x_init_isoc(spca50x);
	}
      out:
	return r;
}

void spca5xx_setFrameDecoder(struct usb_spca50x *spca50x)
{
	int i;
	/* configure the frame detector with default parameters */
	memset(&spca50x->pictsetting, 0, sizeof(struct pictparam));
	spca50x->pictsetting.change = 0x01;
	spca50x->pictsetting.force_rgb = 0;
	spca50x->pictsetting.gamma = 3;
	spca50x->pictsetting.OffRed = 0;
	spca50x->pictsetting.OffBlue = 0;
	spca50x->pictsetting.OffGreen = 0;
	spca50x->pictsetting.GRed = 256;
	spca50x->pictsetting.GBlue = 256;
	spca50x->pictsetting.GGreen = 256;

	for (i = 0; i < SPCA50X_NUMFRAMES; i++) {
		spca50x->frame[i].width = spca50x->width;
		spca50x->frame[i].height = spca50x->height;
		spca50x->frame[i].cameratype = spca50x->cameratype;
		spca50x->frame[i].scanlength = spca50x->width * spca50x->height * 3 / 2;	// assumes 4:2:0 data
		//spca50x->frame[i].depth = 24;
		spca50x->frame[i].depth = 8;

		/* Note: format reflects format of data as returned to
		 * a process, not as read from camera hardware.
		 * This might be a good idea for dumb programs always
		 * assuming the following settings.
		 */
		//spca50x->frame[i].format = VIDEO_PALETTE_RGB24;
		spca50x->frame[i].format = VIDEO_PALETTE_JPEG;
	}
	spca50x->format = VIDEO_PALETTE_JPEG;
}

int spca50x_init_source(struct usb_spca50x *spca50x)
{
	int err_code;

	err_code = zc3xx_init(spca50x);
	if (err_code < 0)
		return -EINVAL;
	spca50x->norme = 0;
	spca50x->channel = 0;
	err_code =
		spca5xx_setMode(spca50x, spca50x->width, spca50x->height, VIDEO_PALETTE_JPEG);
	if (err_code < 0)
		return -EINVAL;
	
	return 0;
}


