
#include "spca_v4l.h"
#include "spca_mem.h"
#include "spca_mode.h"
#include "zc3xx.h"

int spca5xx_do_ioctl(struct inode *inode, struct file *file, unsigned int cmd,
		 void *arg)
{
	struct video_device *vdev = file->private_data;
	struct usb_spca50x *spca50x = video_get_drvdata(vdev);

	PDEBUG(2, "do_IOCtl: 0x%X", cmd);

	if (!spca50x->dev)
		return -EIO;

	switch (cmd) {
	case VIDIOCGCAP:
		{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,22)
			struct video_capability *b = arg;
#else
			struct video_capability j;
			struct video_capability *b = &j;
#endif
			PDEBUG(2, "VIDIOCGCAP %p :", b);

			memset(b, 0, sizeof(struct video_capability));
			/* FIXME: */
			snprintf(b->name, 32, "Z-star Vimicro zc0301p");
			b->type = VID_TYPE_CAPTURE;
			b->channels =
			    ((spca50x->bridge == BRIDGE_SPCA506) ? 8 : 1);
			b->audios = 0;
			b->maxwidth = spca50x->maxwidth;
			b->maxheight = spca50x->maxheight;

			b->minwidth = spca50x->minwidth;
			b->minheight = spca50x->minheight;
			return 0;
		}
	case VIDIOCGCHAN:
		{
			struct video_channel *v = arg;
			/* FIXME:: */
			snprintf(v->name, 32, "ZC301P");
			v->flags = 0;
			v->tuners = 0;
			v->type = VIDEO_TYPE_CAMERA;
			return 0;
		}
	case VIDIOCSCHAN:
		{
			struct video_channel *v = arg;
			/* exclude hardware channel reserved */
			if ((v->channel < 0) || (v->channel > 9)
			    || (v->channel == 4)
			    || (v->channel == 5))
				return -EINVAL;
			return 0;
		}
	case VIDIOCGPICT:
		{
			struct video_picture *p = arg;

			p->depth = spca50x->frame[0].depth;
			p->palette = spca50x->format;
			p->brightness = zc3xx_getbrightness(spca50x) << 8;

			PDEBUG(4, "VIDIOCGPICT: depth=%d, palette=%d", p->depth,
			       p->palette);
			/* default to not allowing any settings to change */
			p->colour = spca50x->colour;
			p->whiteness = spca50x->whiteness;
			p->contrast = spca50x->contrast;
			p->brightness = spca50x->brightness;
			p->hue = spca50x->hue;

			return 0;
		}
	case VIDIOCSPICT:
		{
			int i;
			struct video_picture *p = arg;
			PDEBUG(4, "VIDIOCSPICT");

			if (!spca5xx_get_depth(spca50x, p->palette))
				return -EINVAL;

			if (p->depth < spca50x->min_bpp)
				return -EINVAL;

			spca50x->contrast = p->contrast;
			spca50x->brightness = p->brightness;
			zc3xx_setbrightness(spca50x);
			zc3xx_setcontrast(spca50x);

			spca50x->contrast = p->contrast;
			spca50x->brightness = p->brightness;
			spca50x->colour = p->colour;
			spca50x->hue = p->hue;
			spca50x->whiteness = p->whiteness;

			return 0;
		}
	case VIDIOCGCAPTURE:
		{
			int *vf = arg;

			PDEBUG(4, "VIDIOCGCAPTURE");
			*vf = 0;
			// no subcapture
			return -EINVAL;
		}
	case VIDIOCSCAPTURE:
		{
			struct video_capture *vc = arg;

			if (vc->flags)
				return -EINVAL;
			if (vc->decimation)
				return -EINVAL;

			return -EINVAL;
		}
	case VIDIOCSWIN:
		{
			int result=0;
			struct video_window *vw = arg;

			PDEBUG(3, "VIDIOCSWIN: width=%d, height=%d, flags=%d",
			       vw->width, vw->height, vw->flags);
			if (vw->x)
				return -EINVAL;
			if (vw->y)
				return -EINVAL;
			if (vw->width > (unsigned int)spca50x->maxwidth)
				return -EINVAL;
			if (vw->height > (unsigned int)spca50x->maxheight)
				return -EINVAL;
			if (vw->width < (unsigned int)spca50x->minwidth)
				return -EINVAL;
			if (vw->height < (unsigned int)spca50x->minheight)
				return -EINVAL;

			result = spca5xx_restartMode(spca50x, vw->width,
					vw->height, spca50x->frame[0].format);

#if 0	/* don't need */
			if (result == 0) {
				spca50x->frame[0].width = vw->width;
				spca50x->frame[0].height = vw->height;
			}
#endif

			return result;
		}
	case VIDIOCGWIN:
		{
			struct video_window *vw = arg;

			memset(vw, 0, sizeof(struct video_window));
			vw->x = 0;
			vw->y = 0;
			vw->width = spca50x->frame[0].width;
			vw->height = spca50x->frame[0].height;
			vw->flags = 0;

			PDEBUG(4, "VIDIOCGWIN: %dx%d", vw->width, vw->height);
			return 0;
		}
	case VIDIOCGMBUF:
		{
			struct video_mbuf *vm = arg;
			int i;
			PDEBUG(2, "VIDIOCGMBUF: %p ", vm);
			memset(vm, 0, sizeof(struct video_mbuf));
			vm->size = SPCA50X_NUMFRAMES * MAX_DATA_SIZE;
			vm->frames = SPCA50X_NUMFRAMES;

			for (i = 0; i < SPCA50X_NUMFRAMES; i++) {
				vm->offsets[i] = MAX_DATA_SIZE * i;
			}
			return 0;
		}
	case VIDIOCMCAPTURE:
		{
			int ret, depth;
			struct video_mmap *vm = arg;

			PDEBUG(3, "CMCAPTURE");
			PDEBUG(3, "CM frame: %d, size: %dx%d, format: %d",
			       vm->frame, vm->width, vm->height, vm->format);
			depth = spca5xx_get_depth(spca50x, vm->format);
			if (!depth || depth < spca50x->min_bpp) {
				err("VIDIOCMCAPTURE: invalid format (%d)",
				    vm->format);
				return -EINVAL;
			}

			if ((vm->frame < 0) || (vm->frame > 3)) {
				err("VIDIOCMCAPTURE: invalid frame (%d)",
				    vm->frame);
				return -EINVAL;
			}

			if (vm->width > spca50x->maxwidth
			    || vm->height > spca50x->maxheight) {
				err("VIDIOCMCAPTURE: requested dimensions too big");
				return -EINVAL;
			}
			if (vm->width < spca50x->minwidth
			    || vm->height < spca50x->minheight) {
				err("VIDIOCMCAPTURE: requested dimensions too small");
				return -EINVAL;
			}
			/* 
			 * If we are grabbing the current frame, let it pass
			 */
			/* why should i wait here CSYNCHRO do the job
			   if (spca50x->frame[vm->frame].grabstate == FRAME_GRABBING)
			   {
			   PDEBUG (4, "MCAPTURE: already grabbing");
			   //      ret = wait_event_interruptible (spca50x->wq,
			   ret = wait_event_interruptible (spca50x->frame[vm->frame].wq,
			   (spca50x->frame[vm->frame].
			   grabstate != FRAME_GRABBING));
			   if (ret)
			   return -EINTR;

			   }
			 */
			if ((spca50x->frame[vm->frame].width != vm->width) ||
			    (spca50x->frame[vm->frame].height != vm->height) ||
			    (spca50x->frame[vm->frame].format != vm->format)) {
				ret = spca5xx_restartMode(spca50x, vm->width,
							vm->height, vm->format);
				if (ret < 0)
					return ret;
				spca50x->frame[vm->frame].width = vm->width;
				spca50x->frame[vm->frame].height = vm->height;
				spca50x->frame[vm->frame].format = vm->format;
				spca50x->frame[vm->frame].depth = depth;

			}

			/* Mark it as ready */
			spca50x->frame[vm->frame].grabstate = FRAME_READY;
			return 0;
		}
	case VIDIOCSYNC:
		{
			int ret;
			unsigned int frame = *((unsigned int *)arg);

			PDEBUG(4, "syncing to frame %d, grabstate = %d", frame,
			       spca50x->frame[frame].grabstate);

			switch (spca50x->frame[frame].grabstate) {
			case FRAME_UNUSED:
				return -EINVAL;
			case FRAME_ABORTING:
				return -ENODEV;
			case FRAME_READY:
			case FRAME_GRABBING:
			case FRAME_ERROR:
			     redo:
				if (!spca50x->dev)
					return -EIO;

				ret =
				    wait_event_interruptible(spca50x->
							     frame[frame].wq,
							     (spca50x->
							      frame[frame].
							      grabstate ==
							      FRAME_DONE));
				if (ret)
					return -EINTR;

				PDEBUG(4,
				       "Synch Ready on frame %d, grabstate = %d",
				       frame, spca50x->frame[frame].grabstate);

				if (spca50x->frame[frame].grabstate ==
				    FRAME_ERROR) {
					goto redo;
				}
				/* Fallthrough.
				 * We have waited in state FRAME_GRABBING until it
				 * becomes FRAME_DONE, so now we can move along.
				 */
			case FRAME_DONE:

				/* Release the current frame. This means that it
				 * will be reused as soon as all other frames are
				 * full, so the app better be done with it quickly.
				 * Can this be avoided somehow?
				 */
				spca50x->frame[frame].grabstate = FRAME_UNUSED;
				PDEBUG(4, "Release frame %d state %d\n", frame,
				       spca50x->frame[frame].grabstate);
				break;
			}	/* end switch */

			return 0;
		}
	case VIDIOCGFBUF:
		{
			struct video_buffer *vb = arg;
			memset(vb, 0, sizeof(struct video_buffer));
			vb->base = NULL;	/* frame buffer not supported, not used */
			return 0;
		}
#if 0	/* don't need */
	case SPCAGVIDIOPARAM:
		{
			struct video_param *vp = arg;
			vp->autobright = (__u8) spca50x->autoexpo;
			vp->quality = (__u8) spca50x->qindex;
			vp->time_interval = (__u16) spca50x->dtimes;
			return 0;
		}

	case SPCASVIDIOPARAM:
		{
			struct video_param *vp = arg;
			switch (vp->chg_para) {
			case CHGABRIGHT:
				spca5xx_chgAuto(spca50x, vp->autobright);
				break;
			case CHGQUALITY:
				spca5xx_chgQtable(spca50x, vp->quality);
				break;
			case CHGTINTER:
				spca5xx_chgDtimes(spca50x, vp->time_interval);
				break;
			default:
				return -EINVAL;
				break;
			}
			return 0;
		}
#endif 
	case VIDIOCKEY:
		return 0;
	case VIDIOCCAPTURE:
		return -EINVAL;
	case VIDIOCSFBUF:
		return -EINVAL;
	case VIDIOCGTUNER:
	case VIDIOCSTUNER:
		return -EINVAL;
	case VIDIOCGFREQ:
	case VIDIOCSFREQ:
		return -EINVAL;
	case VIDIOCGAUDIO:
	case VIDIOCSAUDIO:
		return -EINVAL;
	default:
		return -ENOIOCTLCMD;
	}			/* end switch */

	return 0;
}


