#include "zc3xx.h"
#include "hv7131r.h"


#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 6)
#define wait_ms(a) msleep((a))
#endif

#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 11)
#define TimeOut 1000
#else
#define TimeOut HZ
#endif

static void spca5xxRegRead(struct usb_device *dev,
			   __u16 reg,
			   __u16 value,
			   __u16 index, __u8 * buffer, __u16 length)
{
	int rc;
	__u8 *kbuffer = NULL;
	__u16 RegType;
	if (length > 0) {
		kbuffer = (__u8 *) kmalloc(length, GFP_KERNEL);
		memcpy(kbuffer, buffer, length);
	}
	RegType = USB_DIR_IN | USB_TYPE_VENDOR | USB_RECIP_DEVICE;
	rc = usb_control_msg(dev,
			     usb_rcvctrlpipe(dev, 0),
			     reg,
			     RegType,
			     (__u16) value, (__u16) index, kbuffer, length,
			     TimeOut);
	if (length > 0) {
		memcpy(buffer, kbuffer, length);
		kfree(kbuffer);
	}
#if 0
	if (buffer) {
		PDEBUG(5, "reg read: 0x%02X, 0x%02X, 0x%02X, 0x%02X: 0x%04X",
		       RegType, reg, value, index, (int)*buffer);
	} else {
		PDEBUG(5, "reg read: 0x%02X, 0x%02X,0x%02X, 0x%02X ", RegType,
		       reg, value, index);
	}
#endif
	return;
}

static void spca5xxRegWrite(struct usb_device *dev,
			    __u16 reg,
			    __u16 value,
			    __u16 index, __u8 * buffer, __u16 length)
{
	int rc;
	__u16 RegType;
	__u8 *kbuffer = NULL;
	if (length > 0) {
		kbuffer = (__u8 *) kmalloc(length, GFP_KERNEL);
		memcpy(kbuffer, buffer, length);
	}
	RegType = USB_DIR_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE;
	rc = usb_control_msg(dev,
			     usb_sndctrlpipe(dev, 0),
			     reg,
			     RegType,
			     (__u16) value, (__u16) index, kbuffer, length,
			     TimeOut);
	if (length > 0) {
		memcpy(buffer, kbuffer, length);
		kfree(kbuffer);
	}
#if 0
	if (buffer) {
		PDEBUG(5, "reg write: 0x%02X, 0x%02X, 0x%02X, 0x%02X: 0x%04X",
		       RegType, reg, value, index, (int)*buffer);
	} else {
		PDEBUG(5, "reg write: 0x%02X, 0x%02X,0x%02X, 0x%02X ", RegType,
		       reg, value, index);
	}
#endif
	return;
}

static __u8 zcxx3wrSensor[][5] = {
	{0x8b, 0xb3, 0x11, 0x12, 0xff},	// HV7131R
	{0x8b, 0x91, 0x00, 0x15, 0x16},
	{0x8b, 0xe0, 0x14, 0x15, 0x16},
	{0, 0, 0, 0, 0}
};
 __u16 zcxx_i2cRead(struct usb_device *dev, __u8 reg)
{
	__u8 retbyte = 0;
	__u8 retval[] = { 0, 0 };

	spca5xxRegWrite(dev, 0xa0, reg, 0x92, NULL, 0);
	spca5xxRegWrite(dev, 0xa0, 0x02, 0x90, NULL, 0);	// <-read command
	spca5xxRegRead(dev, 0xa1, 0x01, 0x0091, &retbyte, 1);	// read status
	spca5xxRegRead(dev, 0xa1, 0x01, 0x0095, &retval[0], 1);	// read Lowbyte
	spca5xxRegRead(dev, 0xa1, 0x01, 0x0096, &retval[1], 1);	// read Hightbyte
	return ((retval[1] << 8) | retval[0]);
}
 __u8 zcxx_i2cWrite(struct usb_device *dev, __u8 reg, __u8 valL,
			  __u8 valH)
{
	__u8 retbyte = 0;
	spca5xxRegWrite(dev, 0xa0, reg, 0x92, NULL, 0);
	spca5xxRegWrite(dev, 0xa0, valL, 0x93, NULL, 0);
	spca5xxRegWrite(dev, 0xa0, valH, 0x94, NULL, 0);
	spca5xxRegWrite(dev, 0xa0, 0x01, 0x90, NULL, 0);	// <-write command
	spca5xxRegRead(dev, 0xa1, 0x01, 0x0091, &retbyte, 1);	// read status
	return (retbyte);
}

 int zcxx_probeSensor(struct usb_spca50x *spca50x)
{
	int i, j;
	__u8 retbyte = 0;
	__u16 checkword = 0;

/* check 3 wires bus */
	i = 0;
	while (zcxx3wrSensor[i][0]) {
		spca5xxRegWrite(spca50x->dev, 0xa0, 0x02, 0x0010, NULL, 0);
		spca5xxRegRead(spca50x->dev, 0xa1, 0x01, 0x0010, &retbyte, 1);
		spca5xxRegWrite(spca50x->dev, 0xa0, 0x01, 0x0000, NULL, 0);
		spca5xxRegWrite(spca50x->dev, 0xa0, 0x00, 0x0010, NULL, 0);
		spca5xxRegWrite(spca50x->dev, 0xa0, 0x01, 0x0001, NULL, 0);
		spca5xxRegWrite(spca50x->dev, 0xa0, zcxx3wrSensor[i][1],
				zcxx3wrSensor[i][0], NULL, 0);
		spca5xxRegWrite(spca50x->dev, 0xa0, 0x03, 0x0012, NULL, 0);
		spca5xxRegWrite(spca50x->dev, 0xa0, 0x01, 0x0012, NULL, 0);
		spca5xxRegWrite(spca50x->dev, 0xa0, 0x05, 0x0012, NULL, 0);
		for (j = 2; j < 5; j++) {
			if (zcxx3wrSensor[i][j] != 0xff) {
				retbyte =
				    (zcxx_i2cRead
				     (spca50x->dev,
				      zcxx3wrSensor[i][j])) & 0xff;
				PDEBUG(2, "kauri: i=%d j=%d retbyte=%04x", i, j, retbyte);
				if (retbyte != 0) {
					PDEBUG(2, "I'm here ( %d | 0x10)", i);
					return (i | 0x10);
				}
			}
		}
		if (zcxx3wrSensor[i][1] == 0x91) {
			/* check Aurelien Jurgen sensor */
			checkword = 0;
			spca5xxRegWrite(spca50x->dev, 0xa0, 0x02, 0x0010, NULL,
					0);
			spca5xxRegRead(spca50x->dev, 0xa1, 0x01, 0x000b,
				       &retbyte, 1);
			checkword = retbyte << 8;
			spca5xxRegRead(spca50x->dev, 0xa1, 0x01, 0x000a,
				       &retbyte, 1);
			checkword |= retbyte;
			PDEBUG(1, "sensor 3w Vga ??? 0x%04X ",
			       (checkword & 0xffff));
			if (checkword == 0xc001 || checkword == 0xe001) {
				spca50x->chip_revision = checkword;
				spca5xxRegRead(spca50x->dev, 0xa1, 0x01, 0x0010,
					       &retbyte, 1);
				spca5xxRegWrite(spca50x->dev, 0xa0, 0x01,
						0x0000, NULL, 0);
				spca5xxRegWrite(spca50x->dev, 0xa0, 0x0d,
						0x003a, NULL, 0);
				spca5xxRegWrite(spca50x->dev, 0xa0, 0x02,
						0x003b, NULL, 0);
				spca5xxRegWrite(spca50x->dev, 0xa0, 0x00,
						0x0038, NULL, 0);
				PDEBUG(2, "Are you here? ");
				return 0x13;
			}

		}
		i++;
	}
	return (-1);
}

 __u16 zc3xxWriteVector(struct usb_spca50x *spca50x, __u16 data[][3])
{
	struct usb_device *dev = spca50x->dev;
	int err = 0;
	int i = 0;
	__u8 buffread;
	while (data[i][0]) {
		if (data[i][0] == 0xa0) {
			/* write registers */
			spca5xxRegWrite(dev, data[i][0], data[i][1], data[i][2],
					NULL, 0);
		} else {
			/* read status */
			spca5xxRegRead(dev, data[i][0], data[i][1], data[i][2],
				       &buffread, 1);
		}
		i++;
		udelay(1000);
	}

	return err;
}

#define CLAMP(x) (unsigned char)(((x)>0xFF)?0xff:(((x)<1)?1:(x)))

 __u8 Tgamma[16] =
    { 0x13, 0x38, 0x59, 0x79, 0x92, 0xa7, 0xb9, 0xc8, 0xd4, 0xdf, 0xe7,
	0xee, 0xf4, 0xf9, 0xfc, 0xff
};
 __u8 Tgradient[16] =
    { 0x26, 0x22, 0x20, 0x1c, 0x16, 0x13, 0x10, 0x0d, 0x0b, 0x09, 0x07,
	0x06, 0x05, 0x04, 0x03, 0x02
};

 __u16 zc3xx_getbrightness(struct usb_spca50x *spca50x)
{
	spca50x->brightness = 0x80 << 8;
	spca50x->contrast = 0x80 << 8;
	return spca50x->brightness;
}
 __u16 zc3xx_getcontrast(struct usb_spca50x *spca50x)
{

	return spca50x->contrast;
}
 void zc3xx_setbrightness(struct usb_spca50x *spca50x)
{
	__u8 brightness;
	brightness = spca50x->brightness >> 8;
	spca5xxRegWrite(spca50x->dev, 0xa0, brightness, 0x011d, NULL, 0);
	if (brightness < 0x70) {
		spca5xxRegWrite(spca50x->dev, 0xa0, brightness + 0x10, 0x018d,
				NULL, 0);
	} else {
		spca5xxRegWrite(spca50x->dev, 0xa0, 0x80, 0x018d, NULL, 0);
	}

}

 void zc3xx_setcontrast(struct usb_spca50x *spca50x)
{
	__u16 contrast;

	int gm0 = 0;
	int gr0 = 0;
	int index = 0;
	int i;
	/* now get the index of gamma table */
	contrast = zc3xx_getcontrast(spca50x);
	if ((index = contrast >> 13) > 6)
		index = 6;
	PDEBUG(2, "starting new table index %d ", index);
	for (i = 0; i < 16; i++) {
		gm0 = Tgamma[i] * index >> 2;
		gr0 = Tgradient[i] * index >> 2;
		spca5xxRegWrite(spca50x->dev, 0xa0, CLAMP(gm0), 0x0120 + i, NULL, 0);	//brightness
		spca5xxRegWrite(spca50x->dev, 0xa0, CLAMP(gr0), 0x0130 + i, NULL, 0);	// contrast
		//PDEBUG(0,"i %d gamma %d gradient %d",i ,Tgamma[i],Tgradient[i]);
	}
}

 int zc3xx_init(struct usb_spca50x *spca50x)
{
	spca5xxRegWrite(spca50x->dev, 0xa0, 0x01, 0x0000, NULL, 0);
	return 0;
}

 void set_zc3xxVGA(struct usb_spca50x *spca50x)
{
	memset(spca50x->mode_cam, 0x00, TOTMODE * sizeof(struct mwebcam));
	spca50x->mode_cam[VGA].width = 640;
	spca50x->mode_cam[VGA].height = 480;
	spca50x->mode_cam[VGA].t_palette =
	    P_JPEG | P_RAW | P_YUV420 | P_RGB32 | P_RGB24 | P_RGB16;
	spca50x->mode_cam[VGA].pipe = 1023;
	spca50x->mode_cam[VGA].method = 0;
	spca50x->mode_cam[VGA].mode = 0;

	spca50x->mode_cam[SIF].width = 352;
	spca50x->mode_cam[SIF].height = 288;
	spca50x->mode_cam[SIF].t_palette =
	    P_YUV420 | P_RGB32 | P_RGB24 | P_RGB16;
	spca50x->mode_cam[SIF].pipe = 1023;
	spca50x->mode_cam[SIF].method = 1;
	spca50x->mode_cam[SIF].mode = 0;

	spca50x->mode_cam[CIF].width = 320;
	spca50x->mode_cam[CIF].height = 240;
	spca50x->mode_cam[CIF].t_palette = P_JPEG;
	spca50x->mode_cam[CIF].pipe = 1023;
	spca50x->mode_cam[CIF].method = 0;
	spca50x->mode_cam[CIF].mode = 1;

	spca50x->mode_cam[QSIF].width = 176;
	spca50x->mode_cam[QSIF].height = 144;
	spca50x->mode_cam[QSIF].t_palette =
	    P_YUV420 | P_RGB32 | P_RGB24 | P_RGB16;
	spca50x->mode_cam[QSIF].pipe = 1023;
	spca50x->mode_cam[QSIF].method = 1;
	spca50x->mode_cam[QSIF].mode = 1;
}

 int zc3xx_config(struct usb_spca50x *spca50x)
{

	int sensor = 0;
	//__u8 bsensor = 0;
	spca50x->qindex = 1;

	//sensor = zcxx_probeSensor(spca50x);
	sensor=0x11;
	switch (sensor) {
	case -1:
		PDEBUG(0, "Find Sensor UNKNOW_0 force Tas5130");
		spca50x->sensor = SENSOR_TAS5130C;
		set_zc3xxVGA(spca50x);
		break;
	case 0x11:
		PDEBUG(0, "Find Sensor HV7131R(c)");
		spca50x->sensor = SENSOR_HV7131C;
		set_zc3xxVGA(spca50x);
		break;
	};
	return 0;
}
 void zc3xx_setquality(struct usb_spca50x *spca50x)
{
	__u8 quality = 0;
	quality = (spca50x->qindex) & 0xff;
	spca5xxRegWrite(spca50x->dev, 0xa0, quality, 0x0008, NULL, 0);
}
 void zc3xx_setAutobright(struct usb_spca50x *spca50x)
{
	__u8 autoval = 0;
	if (spca50x->autoexpo)
		autoval = 0x42;
	else
		autoval = 0x02;
	spca5xxRegWrite(spca50x->dev, 0xa0, autoval, 0x0180, NULL, 0);

}

 void zc3xx_start(struct usb_spca50x *spca50x)
{
	int err = 0;
	/* Assume start use the good resolution from spca50x->mode */
	//err = zcxx_probeSensor(spca50x);
	switch (spca50x->sensor) {
	case SENSOR_HV7131C:	//HV7131R
	case SENSOR_HV7131R:
		switch(spca50x->width) { 
		case 160: 
			err = zc3xxWriteVector(spca50x, hv7131r_qcif_data);
			break;
		case 176: 
			err = zc3xxWriteVector(spca50x, hv7131r_qsif_data);
			break;
		case 320: 
			err = zc3xxWriteVector(spca50x, hv7131r_cif_data);
			break;
		case 352: 
			err = zc3xxWriteVector(spca50x, hv7131r_sif_data);
			break;
		case 640: 
			err = zc3xxWriteVector(spca50x, hv7131r_vga_data);
			break;
		default:
			break;
		}
	}
//	zc3xx_setbrightness(spca50x);
	zc3xx_setquality(spca50x);
	zc3xx_setAutobright(spca50x);
}

 void zc3xx_stop(struct usb_spca50x *spca50x)
{
// Nothing todo 
}
 void zc3xx_shutdown(struct usb_spca50x *spca50x)
{

	struct usb_device *dev = spca50x->dev;
	__u8 buffread;
	spca5xxRegRead(dev, 0xa1, 0x01, 0x0180, &buffread, 1);
	spca5xxRegWrite(dev, 0xa0, 0x00, 0x0180, NULL, 0);
	spca5xxRegWrite(dev, 0xa0, 0x01, 0x0000, NULL, 0);
}
