/* 
 * This program is grabbing images from zc301p webcam.
 * created by kimsungjo(kauri.kim@gmail.com)
 *
 * This program is under GPL. 
 * */

#include "main.h"
#define LOG_PATH "./rg.log"
#define GMAX 5

FILE *glog = NULL;
int debug = 5;
unsigned char f_log = ZC_PRINT_INFO | ZC_PRINT_ERR | ZC_PRINT_DEBUG;

static unsigned int max_num = 1;	/* take max_num pict. */

static inline check_args(struct zc3xx_cam *cam);
static int option(struct zc3xx_cam *cam, int argc, char **argv);

int main(int argc, char **argv)
{
	int idx = 0;
	struct zc3xx_cam *cam;	/* handle for the grabbing */

	log_init(LOG_PATH);

	cam = (struct zc3xx_cam *)malloc(sizeof(struct zc3xx_cam));
	if (cam == NULL) {
		exit(1);
	}
	bzero(cam, sizeof(struct zc3xx_cam));

	if (option(cam, argc, argv) < 0) {
		PRINT_ERR("failed option");
	}

	check_args(cam);

	if (grab_init(cam) < 0)
		return -1;

	do {
		FILE *fout;
		char name[55];
		int size, len, num;

		size = cam->img_len;
		len = read(cam->cfd, cam->img_buf, size);
		if (len <= 0) {
			printf("v4l read error\n");
			printf("len %d asked %d \n", len, size);
			return 0;
		}
		if (cam->overwrite == 1)
			num = 0;
		else
			num = idx;

		sprintf(name, "%s%d.jpg", cam->filename, num);
		printf("%s\n", name);
		fout = fopen(name, "wb");
		fwrite(cam->img_buf, len, 1, fout);
		fflush(fout);
		fclose(fout);
	} while (idx++ < cam->max_img_num);

	return 0;
}

static inline check_args(struct zc3xx_cam *cam)
{
	if (cam->devpath == NULL) {
		cam->devpath = strdup("/dev/video0");
	}
	if (cam->filename == NULL) {
		cam->filename = strdup("test");
	}
	if (cam->width == 0) {
		cam->width = 320;
		cam->height = 240;
	}
}

static int option(struct zc3xx_cam *cam, int argc, char **argv)
{
	int daemonize = 0;

	/* def value */
	cam->width = 320;
	cam->height = 240;

	while (1) {
		int r = getopt(argc, argv, "d:m:n:f:ho");
		if (r == -1)
			break;

		switch (r) {
		case 'o':
			cam->overwrite = 1;
			break;
		case 'f':
			cam->filename = strdup(optarg);
			break;
		case 'n':
			cam->max_img_num = atoi(optarg);
			if (cam->max_img_num == 0)
				cam->max_img_num = 1;
			break;
		case 'd':
			cam->devpath = strdup(optarg);
			break;
		case 'm':
			if (strlen(optarg) < 2) {
				printf("Unkwon image mode");
			}
			if (!strcmp(optarg, "cif") || !strcmp(optarg, "CIF")) {
				cam->width = 320;
				cam->height = 240;
			} else if (!strcmp(optarg, "vga")
				   || !strcmp(optarg, "VGA")) {
				cam->width = 640;
				cam->height = 480;
			} else {
				fprintf(stderr, "vga=640*480\ncif=320*240\n");
				exit(0);
			}
		case 'h':
			fprintf(stderr, "uage: %s [-w -h -n -f]\n");
			fprintf(stderr, "\t-d: device file path\n");
			fprintf(stderr, "\t-n: grabbing count number\n");
			fprintf(stderr,
				"\t-f: image file name (def:test%%04d.jpg)\n");
			fprintf(stderr, "\t-o: overwrite image file\n");
			break;
		default:
			fprintf(stderr, "Unkown option\n");
			break;
		}
	}

	return 0;
}
