
#include "main.h"

int set_vd_size(struct zc3xx_cam *cam, int width, int height)
{
	int vfd;
	struct video_window *win;

	if (cam == NULL)
		return -1;

	vfd = cam->cfd;
	win = &(cam->videowin);

	if (ioctl(vfd, VIDIOCGWIN, win) < 0) {
		PRINT_ERR("failed VIDIOCGWIN");
		return -1;
	}

	win->width = cam->width = width;
	win->height = cam->height = height;
	win->clipcount = 0;

	if (ioctl(vfd, VIDIOCSWIN, win) < 0) {
		PRINT_ERR("failed VIDIOCSWIN");
		return -1;
	}

	cam->img_len = cam->width * cam->height;
	cam->img_buf =
	    (char *)realloc(cam->img_buf, sizeof(char) * cam->img_len);
	//cam->img_buf = (char*)malloc(sizeof(char)*20000);
	if (cam->img_buf == NULL)
		return -1;
	print_msg(4, "set image buffer size %d", cam->img_len);
	return 0;
}

int grab_init(struct zc3xx_cam *cam)
{
	int vfd;
	struct video_channel vch;
	struct video_picture pict;
	struct video_capability vcap;

	if (cam == NULL) {
		return -1;
	}

	cam->cfd = open(cam->devpath, O_RDWR);
	vfd = cam->cfd;
	if (cam->cfd < 0) {
		PRINT_ERR("failed open %s", cam->devpath);
		return -1;
	}

	if (ioctl(vfd, VIDIOCGCAP, &(cam->videocap)) < 0) {
		PRINT_ERR("failed VIDIOCGCAP");
		return -1;
	}
	cam->videocap.name[strlen(cam->videocap.name) - 1] = '\0';
	PRINT_INFO("Found Camera %s\n", cam->videocap.name);
	PRINT_DEBUG("channels: %d\nmaxwidth=%d\tmaxheight=%d\n"
		    "minwidth=%d\tminheight=%d\n", cam->videocap.channels,
		    cam->videocap.maxwidth, cam->videocap.maxheight,
		    cam->videocap.minwidth, cam->videocap.minheight);

	vch.channel = 1;
	if (ioctl(vfd, VIDIOCGCHAN, &vch) < 0) {
		PRINT_ERR("failed VIDIOGCHAN");
		return -1;
	}
	PRINT_INFO("channel:%d name:%s type:%d\n",
		   vch.channel, vch.name, vch.type);
	//cam->videochan->channel, cam->videochan->name, cam->type);

	if (!(vch.type & VIDEO_TYPE_CAMERA)) {
		PRINT_ERR("NOT VID_TYPE_CAPTURE");
		exit(1);
	}

	if (ioctl(vfd, VIDIOCGPICT, &pict) < 0) {
		perror("failed VIDIOCGPICT");
		exit(1);
	}
	pict.palette = VIDEO_PALETTE_JPEG;
	pict.depth = JPEG_FORMAT_DEPTH;
	if (ioctl(vfd, VIDIOCSPICT, &pict) < 0) {
		PRINT_ERR("failed VIDEOCSPICT");
		exit(1);
	}

	if (set_vd_size(cam, cam->width, cam->height) < 0)
		return -1;

	return 0;
}
