/* This program is under GPL 
 * Copyright: Kim Sungjo <kauri.kim@gmail.com> 
 * */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <string.h>
#include <pthread.h>
#include <SDL/SDL.h>
#include <SDL/SDL_thread.h>
#include <SDL/SDL_audio.h>
#include <SDL/SDL_timer.h>
#include <linux/videodev.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <errno.h>
#include <fcntl.h>
#include <time.h>
#include <sys/time.h>
#include <signal.h>

#include "video.h"
#include "jpeg.h"
#include "print_xxx.h"

/* debug msg option */
int debug=2;	
FILE* glog=NULL;
unsigned char f_log =  ZC_PRINT_ERR | ZC_PRINT_INFO | ZC_PRINT_DEBUG;

int start_sdl(struct zc3xx_cam* cam);

static int 
refresh_screen(unsigned char *src, unsigned char *pict, int width, int height, int size);

static void check_args(struct zc3xx_cam *cam);
static int option(struct zc3xx_cam *cam, int argc, char **argv);

int main(int argc, char** argv) { 
	struct zc3xx_cam* cam;

	log_init("./log.log");

	cam = (struct zc3xx_cam*)malloc(sizeof(struct zc3xx_cam));
	if (cam == NULL) { 
		return -1;
	}
	if (option(cam, argc, argv) < 0) { 
		fprintf(stdout, "wrong argsss.. \n");
		return -1;
	}
	check_args(cam);
	start_sdl(cam);

	if (cam) 
		free(cam);
	return 0;
}

int start_sdl(struct zc3xx_cam* cam) { 
	/*** SDL output screen pointer ***/
	SDL_Surface *pscreen;
	SDL_Event sdlevent;
	unsigned char *p = NULL;
	unsigned char *pp = NULL;
	
	struct pic_struct *pic;
	pthread_t save_pict_t;
	int run=1;	/* sdl window run */

	/* init v4l */
	if (grab_init(cam) < 0) { 
		PRINT_ERR("err grab_init");
		exit(1);
	}

	pic = (struct pic_struct*)malloc(sizeof(struct pic_struct));

	if ((SDL_Init(SDL_INIT_VIDEO | SDL_INIT_TIMER) == -1)) {
		printf("Could not initialize SDL: %s.\n", SDL_GetError());
		exit(-1);
	}

	/* clean up on exit */
	atexit(SDL_Quit);

	PRINT_INFO("SDL initialized.. ");
	
	/* create Window */
	pscreen =
		SDL_SetVideoMode(cam->width, cam->height, 16, SDL_SWSURFACE);
	if (pscreen == NULL) {
		print_msg(0, "[ERR][failed SDL_SetVideoMode][%s]", SDL_GetError());
		exit(1);
	}
	SDL_WM_SetCaption("zc3xx viewer", NULL);

	pic->overwrite = cam->overwrite;
	if (start_save_thread(pic, cam->filename, cam->img_len) < 0) { 
		exit(1);
	}

	/* main loop */
	while (run) {
		unsigned int len=0;
		char* tmp_buf = cam->img_buf;
		memset(tmp_buf, 0x00, cam->img_len);

		len = read(cam->cfd, tmp_buf, cam->img_len);
		if (len < 0) {
			PRINT_ERR("failed read image!!!");
			exit(1);
		}
		//PRINT_DEBUG("read image length: %d", len);

		if (pic->issave) {
			printf("Are you here\n");
			memcpy(pic->buf, tmp_buf, len);
			pic->size = len;
			request_save_img(pic);
		}

		p = pscreen->pixels;
		if (refresh_screen(tmp_buf, p, cam->width, cam->height, (cam->img_len*3)) < 0) { 
			print_msg(0, "[ERR][failed refresh_screen]");
		}
		SDL_UpdateRect(pscreen, 0, 0, 0, 0);
		if (SDL_PollEvent(&sdlevent) == 1) {
			switch (sdlevent.type) {
			case SDL_KEYDOWN:
				switch (sdlevent.key.keysym.sym) {
				case SDLK_g: 
					pic->issave = 1;
					break;
				case SDLK_q:
					printf("\nStop asked\n");
					run = 0;
					break;
				default:
					break;
				}
				break;
			case SDL_QUIT: 
				run=0;
				break;
			}
		}
	}
			
	stop_save_thread(pic);
	grab_release(cam);
	free(pic);
	return 0;
}

#define ADDRESSE(x,y,w) (((y)*(w))+(x))
static void resize16(unsigned char *dst, unsigned char *src, int Wd, int Hd, int Ws,
	      int Hs)
{
	int rx, ry;
	int xscale, yscale;
	int x, y;
	Myrgb24 pixel;
	Myrgb16 *output = (Myrgb16 *) dst;
	Myrgb24 *input = (Myrgb24 *) src;

	xscale = (Ws << 16) / Wd;
	yscale = (Hs << 16) / Hd;
	for (y = 0; y < Hd; y++) {
		for (x = 0; x < Wd; x++) {
			rx = x * xscale >> 16;
			ry = y * yscale >> 16;
			output->blue =
			    input[ADDRESSE((int)rx, (int)ry, Ws)].blue >> 3;
			output->green =
			    input[ADDRESSE((int)rx, (int)ry, Ws)].green >> 2;
			output->red =
			    input[ADDRESSE((int)rx, (int)ry, Ws)].red >> 3;
			output++;
		}
	}

}
/* refresh_screen input all palette to RGB24 -> RGB565 */
static int  
refresh_screen(unsigned char *src, unsigned char *pict, int width, int height, int size)
{
	int i;
	unsigned char *dst = NULL;

	/* for some strange reason tiny_jpegdecode need 1 macroblok line more ?? */
	dst = malloc(width * (height + 8) * 3);
	if (jpeg_decode(&dst, src, &width, &height) < 0 ) { 
		PRINT_ERR("failed jpeg_decode");
		return -1;
	}

	/* rezize16 input rgb24 output rgb565 */
	resize16(pict, dst, width, height, width, height);
	free(dst);
	return 0;
}

static void check_args(struct zc3xx_cam *cam)
{
	if (cam->devpath == NULL) {
		cam->devpath = strdup("/dev/video0");
	}
	if (cam->filename == NULL) {
		cam->filename = strdup("test");
	}
	if (cam->width == 0) {
		cam->width = 320;
		cam->height = 240;
	}
}

static int option(struct zc3xx_cam *cam, int argc, char **argv)
{
	int daemonize = 0;

	/* def value */
	cam->width = 320;
	cam->height = 240;

	while (1) {
		int r = getopt(argc, argv, "d:m:n:f:hl:o");
		if (r == -1)
			break;

		switch (r) {
		case 'h':
			fprintf(stdout, "******* Excute Command ********\n");
			fprintf(stdout, "uage: %s [-d -m -f -o -h]\n", argv[0]);
			fprintf(stdout, "-d: device file path\n");
			fprintf(stdout, "-m: indicate image mode [vga:cif:scif]\n");
			fprintf(stdout, "-f: shapshoot image file name (def:test%%04d.jpg)\n");
			fprintf(stdout, "-o: overwrite image file\n");
			fprintf(stdout, "-l: log level for debug [default=0]\n");
			fprintf(stdout, "******* SDL Command ********\n");
			fprintf(stdout, "g: snapshoot\n");
			fprintf(stdout, "q: quit program\n");
			exit(0);
		case 'l': 
			debug=atoi(optarg);
		case 'o':
			cam->overwrite = 1;
			break;
		case 'f':
			cam->filename = strdup(optarg);
			break;
		case 'n':
			cam->max_img_num = atoi(optarg);
			if (cam->max_img_num == 0)
				cam->max_img_num = 1;
			break;
		case 'd':
			cam->devpath = strdup(optarg);
			break;
		case 'm':
			if (strlen(optarg) < 2) {
				printf("Unkwon image mode");
			}
			if (!strcmp(optarg, "cif") || !strcmp(optarg, "CIF")) {
				cam->width = 320;
				cam->height = 240;
			} else if (!strcmp(optarg, "vga")
				   || !strcmp(optarg, "VGA")) {
				cam->width = 640;
				cam->height = 480;
			} else {
				fprintf(stdout, "vga=640*480\ncif=320*240\n");
				exit(0);
			}
		default:
			fprintf(stdout, "Unkown option\n");
			break;
		}
	}

	return 0;
}
