
#ifndef _UTIL_H_
#define _UTIL_H_

#include <errno.h>
#include <stdio.h>

/*
 * 0: no message
 * 1: err message
 * 2: info message
 * 3: debug message
 */

#define ZC_PRINT_ERR 1
#define ZC_PRINT_INFO 1 << 1
#define ZC_PRINT_DEBUG 1 << 2

extern int debug;	/* main.c */
extern FILE* glog;	/* main.c */

extern unsigned char f_log; /* ZC_PRINT_ERR | ZC_LOG_INFO | ZC_LOG_DEBUG */

#ifdef ZC_LOG_FILE
#define log_init(logpath)	\
	do {							\
		glog = fopen(logpath, "a");	\
		if (glog == NULL){				\
			fprintf(stderr, "err fopen");	\
			exit(-1);				\
		}						\
	}while(0)

#define log_release()	fclose(glog)
#define print_msg( level, fmt, args...) 	\
	do {					\
		if (level > debug) break;	\
		fprintf(glog, "[%d] [%s:%s:%d] " fmt "\n", 	\
			level, __FILE__, __PRETTY_FUNCTION__,\
			__LINE__, ## args);	\
	} while(0)

#else 
#define log_init(logpath)	
#define log_release()
#define print_msg( level, fmt, args...) 	\
	do {					\
		if (level > debug) break;	\
		fprintf(stdout, "[%d] [%s:%s:%d] " fmt "\n", 	\
			level, __FILE__, __PRETTY_FUNCTION__,\
			__LINE__, ## args);	\
	} while(0)

#endif

#ifdef MDEBUG
#define PRINT_DEBUG(fmt, args...) if (ZC_PRINT_DEBUG & f_log)  print_msg(4, fmt, ## args)
#else
#define PRINT_DEBUG(fmt, args...)
#endif

/*
 * Information message  θ 
 */
#ifdef INFO
#define PRINT_INFO(fmt, args...) if (f_log & ZC_PRINT_INFO) print_msg(2, fmt, ## args)
#else
#define PRINT_INFO(fmt, args...)
#endif

/*
 *  ޼  ũ
 */
#ifdef ERR
#define PRINT_ERR(fmt, args...)  if (f_log & ZC_PRINT_ERR) \
			print_msg(1, fmt " [errno %d - %s] " , ## args, errno, strerror(errno))
#else
#define PRINT_ERR(fmt, args...)
#endif

#endif				//_UTIL_H_
