/****************************************************************************
#	 	spcaview: Sdl video recorder and viewer with sound.         #
#This package work with the spca5xx based webcam with the raw jpeg feature. #
#All the decoding is in user space with the help of libjpeg.                #
#.                                                                          #
# 		Copyright (C) 2003 2004 2005 Michel Xhaard                  #
#                                                                           #
# This program is free software; you can redistribute it and/or modify      #
# it under the terms of the GNU General Public License as published by      #
# the Free Software Foundation; either version 2 of the License, or         #
# (at your option) any later version.                                       #
#                                                                           #
# This program is distributed in the hope that it will be useful,           #
# but WITHOUT ANY WARRANTY; without even the implied warranty of            #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             #
# GNU General Public License for more details.                              #
#                                                                           #
# You should have received a copy of the GNU General Public License         #
# along with this program; if not, write to the Free Software               #
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA #
#                                                                           #
****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <string.h>
#include <pthread.h>
#include <SDL/SDL.h>
#include <SDL/SDL_thread.h>
#include <SDL/SDL_audio.h>
#include <SDL/SDL_timer.h>
#include <linux/videodev.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <errno.h>
#include <fcntl.h>
#include <time.h>
#include <sys/time.h>
#include <signal.h>
#include "SDL_audioin.h"

#include "jconfig.h"
#include "avilib.h"
#include "utils.h"
#include "version.h"

#define WAVE_AUDIO_PCM 1
#define CHANNELS 1
#define AUDIO_SAMPLERATE		22050	// 44100 // 11025 // 8000 // 22050
#define AUDIO_SAMPLES			512	// 1024
#define AUDIO_FORMAT			AUDIO_S16
#define AUDIOSIZE 2
#define NUM_CHUNCK 64		// 256 a large ring buffer
#define MAXBUFFERSIZE (AUDIO_SAMPLES*AUDIOSIZE*NUM_CHUNCK)
#define HELPBUFFERSIZE (AUDIO_SAMPLES*AUDIOSIZE)

// INIT_BRIGHT is the initial brightness level
// BRIGHTMEAN is the goal mean brightness for the 
// auto adjustment.
// SPRING_CONSTANT determines the restoring force
// for the auto adjust
// BRIGHT_WINDOW is the window around the goal mean
// in which no adjustment is to occur
#define V4L_BRIGHT_MIN 0
#define V4L_BRIGHT_MAX 50000	//65000
#define BRIGHTMEAN 128		//145
#define SPRING_CONSTANT 25
#define BRIGHTWINDOW 10

/* pictFlag should be set by the hardware or the timer */
/* call_back function need to reset the flag */
static int totmean;
int interval = 0;
volatile int pictFlag = 0;
static int Oneshoot = 1;
static int numshoot = 0;
static int modshoot = 0;
static videoOk = 0;

/* picture structure */
typedef struct pictstruct {
	unsigned char *data;
	int sizeIn;
	int width;
	int height;
	int formatIn;
	int mode;
	avi_t *out_fd;
	pthread_mutex_t mutex;
	pthread_cond_t cond;
} pictstruct;

void *waitandshoot(void *ptr);

/* sound buffer */
struct Rbuffer {
	Uint8 buffer[MAXBUFFERSIZE];
	Uint8 helpbuffer[HELPBUFFERSIZE];
	int ptread;
	int ptwrite;
};

static Uint32 callback_timer(Uint32 interval)
{
	pictFlag = 1;
	printf("TIMER ON !!!!\n");
	return interval;
}

static int setVideoPict(struct video_picture *videopict, int fd);

static int isSpcaChip(const char *BridgeName);
/* return Bridge otherwhise -1 */
static int getStreamId(const char *BridgeName);
/* return Stream_id otherwhise -1 */
static int probeSize(const char *BridgeName, int *width, int *height);
/* return 1->ok otherwhise -1 */
void
resize(unsigned char *dst, unsigned char *src, int Wd, int Hd, int Ws, int Hs);

static void
refresh_screen(unsigned char *src, unsigned char *pict, int format, int width,
	       int height, int owidth, int oheight, int size, int autobright);

int spcaGrab(char *outputfile, char fourcc[4], const char *videodevice,
	     int image_width, int image_height, int format, int owidth,
	     int oheight, int grabMethod, int videoOn, int audioout,
	     int videocomp, int autobright, int statOn, int decodeOn);

/* called spcaGrab */
void *waitandshoot(void *ptr)
{
	pictstruct *mypict = (pictstruct *) ptr;
	int width = mypict->width;
	int height = mypict->height;
	int format = mypict->formatIn;
	int size = mypict->sizeIn;
	int mode = mypict->mode;
	avi_t *fd = mypict->out_fd;
	int status;
	while (1) {
		status = pthread_mutex_lock(&mypict->mutex);
		if (status != 0) {
			fprintf(stderr, "Lock error!\n");
			exit(-1);
		}
		while (pictFlag == 0) {
			/* Set predicate */
			status = pthread_cond_wait(&mypict->cond, &mypict->mutex);
			if (status != 0) {
				fprintf(stderr, "Wait error!\n");
				exit(-1);
			}
		}
		pictFlag = 0;
		getJpegPicture(mypict->data, width, height, format, size, mode,
			       fd);
		if (modshoot)
			if (!(--numshoot))
				Oneshoot = 0;
		status = pthread_mutex_unlock(&mypict->mutex);
		if (status != 0) {
			fprintf(stderr, "Unlock error!\n");
			exit(-1);
		}
    /******* EXIT the MONITOR ************/
// printf("Condition was accept, data.value=%d\n",pictFlag);

	}
	return NULL;

}

int main(int argc, char *argv[])
{
	/* Timing value used for statistic */

	const char *videodevice = NULL;
	/* default mmap */
	int grabMethod = 0;

	//int format = VIDEO_PALETTE_YUV420P;
	int format = VIDEO_PALETTE_JPEG;	/* kauri:OK */
	/******** output screen pointer ***/

	int image_width = IMAGE_WIDTH;
	int image_height = IMAGE_HEIGHT;
	int owidth = 0;
	int oheight = 0;
	/**********************************/
	/*        avi parametres          */
	char *inputfile = NULL;
	char *outputfile = NULL;
	char fourcc[4] = "MJPG";

	char *sizestring = NULL;
	int use_libjpeg = 1;
	char *separateur;
	char *mode = NULL;
	/*********************************/
	/*          Starting Flags       */
	int i;
	int videoOn = 1;
	int decodeOn = 1;
	int statOn = 0;
	int audioout = 0;
	int videocomp = 0;
	int channel = 0;
	int norme = 0;
	int autobright = 0;
	/*********************************/
	char *AdIpPort;
	char AdIp[] = "000.000.000.000";
	unsigned short ports = 0;

	/*********************************/
	SPCASTATE funct;

	/* init default bytes per pixels for VIDEO_PALETTE_RAW_JPEG     */
	/* FIXME bpp is used as byte per pixel for the ouput screen     */
	/* we need also a bpp_in for the input stream as spcaview       */
	/* can convert stream That will be a good idea to have 2 struct */
	/* with all global data one for input the other for output      */
	bpp = 3;	/* check */
	funct = GRABBER;
	printf(" %s \n", version);

	switch (funct) {
	case GRABBER:
		{
			/* spcaview grabber */
			spcaGrab(outputfile, fourcc, videodevice, image_width,
				 image_height, format, owidth, oheight,
				 grabMethod, videoOn, audioout, videocomp,
				 autobright, statOn, decodeOn);
		}
		break;
	}

	exit(0);
}

int spcaGrab(char *outputfile, char fourcc[4], const char *videodevice,
	     int image_width, int image_height, int format, int owidth,
	     int oheight, int grabMethod, int videoOn, int audioout,
	     int videocomp, int autobright, int statOn, int decodeOn)
{

	/* Timing value used for statistic */
	Uint32 total_decode_time = 0;
	Uint32 time = 0;
	Uint32 prevtime = 0;
	Uint32 synctime = 0;
	Uint32 decodetime = 0;
	double average_decode_time = 0;
	Uint32 intime = 0;
	Uint32 pictime = 0;
	Uint32 delaytime = 0;
	Uint32 compresstime = 0;
	/**********************************/
	/*        video structures       */
	int fd;
	/* default mmap */
	int i, j, k, nframes = 2000, f, status;
	struct video_mmap vmmap;
	struct video_capability videocap;
	int mmapsize;
	struct video_mbuf videombuf;
	struct video_picture videopict;
	struct video_window videowin;
	struct video_channel videochan;
	struct video_param videoparam;
	int norme = 0;
	int channel = 0;
	/******** output screen pointer ***/
	SDL_Surface *pscreen;
	SDL_Event sdlevent;
	unsigned char *p = NULL;
	unsigned char *pp = NULL;
	/**********************************/
	/*        avi parametres          */
	avi_t *out_fd;
	unsigned char *pFramebuffer;
	unsigned char *tmp = NULL;	// don't forget the Till lesson
	unsigned char *dpshDest = NULL;
	u_int8_t *jpegData = NULL;	// Till lesson a good pointer is a pointer NULL
	int jpegSize;
	int ff;
	int framecount = 0;
	/*********************************/
	/*          Starting Flags       */
	int run = 1;
	int quit = 1;

	int initAudio = 0;	//flag start audio
	int method = 1;
	int streamid;
	int isVideoChannel = 1;
	int frame_size = 0;
	int len = 0;

	/*********************************/
	/* data for SDL_audioin && SDL_audio */
	SDL_AudioSpec spec, result;
	SDL_AudioSpec expect;
	struct Rbuffer RingBuffer;
	int retry = 100;
	int ptread;
	int ptwrite;
	int err = 0;
	int bytes_per_read = 0;
	int testbpp = 16;
	/*********************************/
	pictstruct mypict;
	pthread_t waitandshoot_id;
	int wstatus;
	/*********************************/
	RingBuffer.ptread = 0;
	RingBuffer.ptwrite = 0;
	/* spcaview grabber */
	printf("Initializing SDL.\n");

	/* Initialize defaults, Video and Audio */
	if ((SDL_Init(SDL_INIT_VIDEO | SDL_INIT_TIMER) == -1)) {
		printf("Could not initialize SDL: %s.\n", SDL_GetError());
		exit(-1);
	}

	/* Clean up on exit */
	atexit(SDL_Quit);
	if (!owidth || !oheight) {
		owidth = image_width;
		oheight = image_height;
	}
	printf("SDL initialized.\n");
	/* validate parameters */
	printf("[check] bpp %d format %d\n", bpp, format);

	if (videodevice == NULL || *videodevice == 0) {
		videodevice = "/dev/video0";
	}
	printf("Using video device %s.\n", videodevice);
	printf("Initializing v4l.\n");

	//v4l init
	if ((fd = open(videodevice, O_RDWR)) == -1) {
		perror("ERROR opening V4L interface \n");
		exit(1);
	}

	printf("**************** PROBING CAMERA *********************\n");
	if (ioctl(fd, VIDIOCGCAP, &videocap) == -1) {
		printf("wrong device\n");
		exit(1);
	}

	printf("Camera found: %s \n", videocap.name);

	if (ioctl(fd, VIDIOCGCHAN, &videochan) == -1) {
		printf("Hmm did not support Video_channel\n");
		isVideoChannel = 0;
	}
	if (isVideoChannel) {
		videochan.norm = norme;
		videochan.channel = channel;
		if (ioctl(fd, VIDIOCSCHAN, &videochan) == -1) {
			printf("ERROR setting channel and norme \n");
			exit(1);
		}
		/************ just to be sure *************/
		if (ioctl(fd, VIDIOCGCHAN, &videochan) == -1) {
			printf("wrong device\n");
			exit(1);
		}
		printf("Bridge found: %s \n", videochan.name);
		streamid = getStreamId(videochan.name);

		if (streamid >= 0) {
		} else {
			printf("StreamId: %d Unknow Camera\n", streamid);
		}
		/* test jpeg capability if not and jpeg ask without raw data 
		   set default format to YUVP */
		if ((format == VIDEO_PALETTE_RAW_JPEG
		     || format == VIDEO_PALETTE_JPEG) && streamid != JPEG
		    && videoOn) {
			printf
			    ("Camera unable to stream in JPEG mode switch to YUV420P\n");
			format = VIDEO_PALETTE_YUV420P;
		}
		if (probeSize(videochan.name, &image_width, &image_height) < 0)
			printf("unable to probe size !!\n");
	}
	printf("*****************************************************\n");
	/* Init grab method mmap */
	if (grabMethod) {
		printf(" grabbing method default MMAP asked \n");
		// MMAP VIDEO acquisition
		memset(&videombuf, 0, sizeof(videombuf));
		if (ioctl(fd, VIDIOCGMBUF, &videombuf) < 0) {
			perror(" init VIDIOCGMBUF FAILED\n");
		}
		printf
		    ("VIDIOCGMBUF size %d  frames %d  offets[0]=%d offsets[1]=%d\n",
		     videombuf.size, videombuf.frames, videombuf.offsets[0],
		     videombuf.offsets[1]);

		pFramebuffer =
		    (unsigned char *)mmap(0, videombuf.size,
					  PROT_READ | PROT_WRITE,
					  MAP_SHARED, fd, 0);
		mmapsize = videombuf.size;
		vmmap.height = image_height;
		vmmap.width = image_width;
		vmmap.format = format;
		for (f = 0; f < videombuf.frames; f++) {
			vmmap.frame = f;
			if (ioctl(fd, VIDIOCMCAPTURE, &vmmap)) {
				perror("cmcapture");
			}
		}
		vmmap.frame = 0;
	} else {
		/* read method */
		printf(" grabbing method READ asked \n");
		if (ioctl(fd, VIDIOCGWIN, &videowin) < 0)
			perror("VIDIOCGWIN failed \n");
		videowin.height = image_height;
		videowin.width = image_width;
		if (ioctl(fd, VIDIOCSWIN, &videowin) < 0)
			perror("VIDIOCSWIN failed \n");
		printf("VIDIOCGWIN height %d  width %d \n",
		       videowin.height, videowin.width);

	}
	switch (format) {
	case VIDEO_PALETTE_JPEG:{
			frame_size = image_width * image_height;
		}
		break;
	case VIDEO_PALETTE_RAW_JPEG:{
			frame_size = image_width * image_height * 3;
		}
		break;
	case VIDEO_PALETTE_YUV420P:{
			frame_size = (image_width * image_height * 3) >> 1;
		}
		break;
	case VIDEO_PALETTE_RGB565:
	case VIDEO_PALETTE_RGB24:
	case VIDEO_PALETTE_RGB32:{
			frame_size = image_width * image_height * bpp;
		}
		break;
	default:
		break;

	}
	/* struct video_picture VIDIOCGPICT VIDIOCSPICT */
	if (ioctl(fd, VIDIOCGPICT, &videopict) < 0) {
		perror("Couldnt get videopict params with VIDIOCGPICT\n");
	}	/* depth=8: format= */
	printf("VIDIOCGPICT\n");
	printf("brightnes=%d hue=%d color=%d contrast=%d whiteness=%d \n",
	       videopict.brightness, videopict.hue, videopict.colour,
	       videopict.contrast, videopict.whiteness);
	printf("depth=%d palette=%d\n", videopict.depth, videopict.palette);

	videopict.palette = format;
	//videopict.depth = bpp * 8;
	videopict.depth =  8;	/* jpeg = 8 */
	//videopict.brightness = INIT_BRIGHT;
	sleep(1);
	setVideoPict(&videopict, fd);	

	/*
	 * Initialize the display 
	 */
	if (decodeOn && videoOn) {	/* Display data */
		printf("decodeOn:%d videoOn:%d\n", decodeOn, videoOn);
		printf("Do you see window? \n");
		/* SDL_RESIZABLE  ߰ */
		pscreen =
		    SDL_SetVideoMode(owidth, oheight, testbpp, SDL_SWSURFACE);
		if (pscreen == NULL) {
			printf("Couldn't set %d*%dx%d video mode: %s\n",
			       owidth, oheight, 3 * 8, SDL_GetError());
			exit(1);
		}
		printf("Do you see window? \n");
		SDL_WM_SetCaption("zc3xx viwer", NULL);
	}
	printf("\n");
	time = SDL_GetTicks();
	prevtime = time;

	/* If need open the avi file and alloc jpegbuffer */
	if (outputfile) {
		jpegData = malloc(frame_size);
		dpshDest = malloc(frame_size);
		if ((out_fd = AVI_open_output_file(outputfile)) == NULL) {
			printf("cannot open write file ? \n");
			exit(1);
		}
		if (interval) {
			/* picture goes in avi */
			AVI_set_video(out_fd, image_width, image_height, 1, "MJPG");
			mypict.mode = AVIPICT;
			mypict.out_fd = out_fd;
		} else {
			AVI_set_video(out_fd, image_width, image_height, 20,
				      fourcc);
		}
	} /* end handling outputfile */

	/* Allocate tmp buffer for one frame. */
	tmp = (unsigned char *)malloc(frame_size);
	/* laugth the picture thread */
	mypict.data = malloc(frame_size);
	mypict.sizeIn = frame_size;
	mypict.width = image_width;
	mypict.height = image_height;
	mypict.formatIn = format;
	pthread_mutex_init(&mypict.mutex, NULL);
	pthread_cond_init(&mypict.cond, NULL);
	/* kauri:check */
	wstatus =
	    pthread_create(&waitandshoot_id, NULL, (void *)waitandshoot,
			   &mypict);
	if (wstatus != 0) {
		fprintf(stderr, "thread shoot Create error!\n");
		exit(-1);
	}

	if (interval && videoOn) {
		/* allocate the buffer for photos */
		if (outputfile) {
			mypict.mode = AVIPICT;
		} else {
			mypict.mode = PICTURE;
		}
		/* ͹  ̹ grabѴ. */
		SDL_SetTimer((Uint32) interval, callback_timer);
	}
	i = 0;

	/* main loop */
	while (run && Oneshoot) {
		memset(tmp, 0x00, frame_size);
		intime = SDL_GetTicks();
		pictime = intime - delaytime;
		delaytime = intime;
		/* Try to synchronize sound with frame rate */
		if (initAudio && i > 9) {
			initAudio = 0;
			SDL_PauseAudioIn(0);	// start record
		}
		/* compute bytes sound */
		if (pictime < 100) {
			bytes_per_read =
			    ((AUDIO_SAMPLERATE / 1000) * 2 * pictime);
		}
		i++;
		if (grabMethod) {
			ff = vmmap.frame;
			if (ioctl(fd, VIDIOCSYNC, &ff) < 0) {
				perror("cvsync err\n");
			}
			vmmap.frame = ff;
			memcpy(tmp,
			       pFramebuffer +
			       videombuf.offsets[vmmap.frame], frame_size);
			if ((status = ioctl(fd, VIDIOCMCAPTURE, &vmmap)) < 0) {
				perror("cmcapture");
				printf(">>cmcapture err %d\n", status);
			}
			vmmap.frame = (vmmap.frame + 1) % videombuf.frames;
		} else {
			/* 츮 ˰ ִ... 츮 ϴ  ϼ Ǿٴ !!! */
			/* read method */
			len = read(fd, tmp, frame_size);
			printf ("len %d asked %d \n",len,frame_size);
		}
		synctime = SDL_GetTicks();
		/* !! */
		/*here the frame is in tmp ready to used */
		if (pictFlag) { /* ̹  ǳ? */
			wstatus = pthread_mutex_lock(&mypict.mutex);
			if (wstatus != 0) {
				fprintf(stderr, "Lock error!\n");
				exit(-1);
			}
			memcpy(mypict.data, tmp, frame_size);
			//printf("COND OK !!\n");
			/*   ;) */
			/* ƹ   ׳ ϰų Ѹȴٳ ģ  */
			wstatus = pthread_cond_signal(&mypict.cond);
			if (wstatus != 0) {
				fprintf(stderr, "Signal error!\n");
				exit(-1);
			}
			wstatus = pthread_mutex_unlock(&mypict.mutex);
			if (wstatus != 0) {
				fprintf(stderr, "Unlock error!\n");
				exit(-1);
			}
		}
		if ((outputfile) && (i > 10) && !interval) {
			/* Output video frame with the good format */
			switch (format) {
				case VIDEO_PALETTE_JPEG:{	
					/* ̰ 츮, ݾ   */
					/* ٸ   ڰ ģ  */
					int fs = get_jpegsize(tmp, frame_size);
					if (AVI_write_frame(out_fd, (unsigned char *)
							    tmp, fs) < 0)
						printf ("write error on avi out \n");
					method = 0;
				}
				break;
			/* 3 */	
			default:
				break;
			}
			/*  ұ? 
			 * ߿ ʿϴٰ Ǹ    ߰ ڰ   ž ȣȣ */
			/* 3 */
		}

		compresstime = SDL_GetTicks();

		/* kauri: Don't Need */
		if (decodeOn && videoOn && (i > 10)) {	/* Display data */
			p = pp = pscreen->pixels;

			refresh_screen(tmp, p, format,
				       image_width, image_height, owidth,
				       oheight,
				       image_width * image_height * bpp,
				       autobright);
			decodetime = SDL_GetTicks();

			SDL_UpdateRect(pscreen, 0, 0, 0, 0);	//update the entire screen      
		} else {
			decodetime = SDL_GetTicks();
		}

		/* increment the real frame count and update statistics */
		framecount++;

		total_decode_time += decodetime - synctime;
		average_decode_time = total_decode_time / framecount;

		if (!autobright && statOn) {
// \r
			printf
			    ("frames:%04d pict:%03dms synch:%03dms write/comp:%03dms decode:%03dms display:%03dms \n",
			     framecount, pictime, synctime - intime,
			     compresstime - synctime, decodetime - compresstime,
			     SDL_GetTicks() - decodetime);
			fflush(stdout);
		}

		if (SDL_PollEvent(&sdlevent) == 1) {
			switch (sdlevent.type) {
			case SDL_KEYDOWN:
				switch (sdlevent.key.keysym.sym) {
				case SDLK_b:
					videopict.brightness -= 0x100;
					setVideoPict(&videopict, fd);
					break;
				case SDLK_n:
					videopict.brightness += 0x100;
					setVideoPict(&videopict, fd);
					break;
				case SDLK_c:
					videopict.colour -= 0x200;	// = saturation ?
					setVideoPict(&videopict, fd);
					break;
				case SDLK_s:
					/* should get the next frame available */
					pictFlag = 1;
					break;
				case SDLK_v:
					videopict.colour += 0x200;	// = saturation ?
					setVideoPict(&videopict, fd);
					break;
				case SDLK_w:
					videopict.contrast -= 0x200;	// -contrast
					setVideoPict(&videopict, fd);
					break;
				case SDLK_x:
					videopict.contrast += 0x200;	// +contrast
					setVideoPict(&videopict, fd);
					break;
				case SDLK_q:
					printf("\nStop asked\n");
					run = 0;
					break;
				default:

					break;
				}
				break;
			case SDL_QUIT:
				run = 0;
				break;
			}
		}

	}

	if (audioout) {
		SDL_CloseAudioIn();	//stop record 
		printf("free sound buffer\n");

	}
	if (interval && decodeOn && videoOn) {
		// set_timer(interval);
		SDL_SetTimer((Uint32) 0, NULL);	
	}
	time = SDL_GetTicks() - time;
	printf("\nUsed %dms for %d images => %dms/image %dfps.\n",
	       time, framecount, time / framecount, framecount * 1000 / time);

	printf("Quiting SDL.\n");

	printf("Decoded frames:%d Average decode time: %f\n",
	       framecount, average_decode_time);
	if (grabMethod) {
		printf("unmapping\n");
		munmap(pFramebuffer, mmapsize);
	}
	printf("closing\n");
	close(fd);
	printf("closed\n");
	if (outputfile) {
		if (!interval)
			out_fd->fps = (double)framecount *1000 / time;

		AVI_close(out_fd);
		printf("close avi\n");
		if (audioout) {

			SDL_QuitAudioIn();
			printf("free sound buffer\n");

		}
		free(jpegData);
		free(dpshDest);
	}

	pthread_cancel(waitandshoot_id);
	pthread_join(waitandshoot_id, NULL);
	pthread_cond_destroy(&mypict.cond);
	pthread_mutex_destroy(&mypict.mutex);
	if (mypict.data)
		free(mypict.data);
	printf("Destroy Picture thread ...\n");

	/* Shutdown all subsystems */
	free(tmp);
	//close_libjpeg_decoder ();
	printf("Quiting....\n");
	SDL_Quit();
	return 0;
}


static int setVideoPict(struct video_picture *videopict, int fd)
{
	if (ioctl(fd, VIDIOCSPICT, videopict) < 0) {
		perror("Couldnt get videopict params with VIDIOCSPICT\n");
		return -1;
	}
	printf("VIDIOCSPICT\n");
	printf("brightness=%d hue=%d color=%d contrast=%d whiteness=%d \n",
	       videopict->brightness, videopict->hue,
	       videopict->colour, videopict->contrast, videopict->whiteness);
	printf("depth=%d palette=%d \n", videopict->depth, videopict->palette);

	return 0;
}


#define ADDRESSE(x,y,w) (((y)*(w))+(x))
void resize16(unsigned char *dst, unsigned char *src, int Wd, int Hd, int Ws,
	      int Hs)
{
	int rx, ry;
	int xscale, yscale;
	int x, y;
	Myrgb24 pixel;
	Myrgb16 *output = (Myrgb16 *) dst;
	Myrgb24 *input = (Myrgb24 *) src;

	xscale = (Ws << 16) / Wd;
	yscale = (Hs << 16) / Hd;
	for (y = 0; y < Hd; y++) {
		for (x = 0; x < Wd; x++) {
			rx = x * xscale >> 16;
			ry = y * yscale >> 16;
			output->blue =
			    input[ADDRESSE((int)rx, (int)ry, Ws)].blue >> 3;
			output->green =
			    input[ADDRESSE((int)rx, (int)ry, Ws)].green >> 2;
			output->red =
			    input[ADDRESSE((int)rx, (int)ry, Ws)].red >> 3;
			output++;
		}
	}

}
void resize(unsigned char *dst, unsigned char *src, int Wd, int Hd, int Ws,
	    int Hs)
{
	int rx, ry;
	int xscale, yscale;
	int x, y;
	Myrgb24 pixel;
	Myrgb24 *output = (Myrgb24 *) dst;
	Myrgb24 *input = (Myrgb24 *) src;

	xscale = (Ws << 16) / Wd;
	yscale = (Hs << 16) / Hd;
	for (y = 0; y < Hd; y++) {
		for (x = 0; x < Wd; x++) {
			rx = x * xscale >> 16;
			ry = y * yscale >> 16;
			memcpy(output++, &input[ADDRESSE((int)rx, (int)ry, Ws)],
			       sizeof(Myrgb24));
		}
	}

}

/* refresh_screen input all palette to RGB24 -> RGB565 */
static void
refresh_screen(unsigned char *src, unsigned char *pict, int format, int width,
	       int height, int owidth, int oheight, int size, int autobright)
{
	unsigned int *lpix;
	unsigned short *pix;
	int i;
	int intwidth = width;
	int intheight = height;
	unsigned char *dst = NULL;
/* in case VIDEO_PALETTE_RAW_JPEG nothing todo */
	if (format == VIDEO_PALETTE_RAW_JPEG)
		return;
	/* for some strange reason tiny_jpegdecode need 1 macroblok line more ?? */
	dst = malloc(width * (height + 8) * 3);
	switch (format) {
		/* FIXME: ߷µǴ  RGB565ΰ */
	case VIDEO_PALETTE_JPEG:{
			//libjpeg_decode (dst, src, size, width, height);
			
			jpeg_decode(&dst, src, &intwidth, &intheight);
		}
		break;
	case VIDEO_PALETTE_YUV420P:{
			// uncompressed data yuv420P decoder in module
			// equalize (src, width, height, 0);
			YUV420toRGB((unsigned char *)src,
				    (unsigned char *)dst, width, height, 0, 0);

		}
		break;
	case VIDEO_PALETTE_RGB565:
		pix = (unsigned short *)src;
		for (i = 0; i < ((width * height * 3) - 3); i += 3) {
			dst[i] = (*pix & 0x001F) << 3;
			dst[i + 1] = (*pix & 0x07E0) >> 3;
			dst[i + 2] = (*pix & 0xF800) >> 8;
			pix++;
		}
		break;
	case VIDEO_PALETTE_RGB32:
		lpix = (unsigned int *)src;
		for (i = 0; i < ((width * height * 3) - 3); i += 3) {
			dst[i] = (*lpix & 0x000000FF);
			dst[i + 1] = (*lpix & 0x0000FF00) >> 8;
			dst[i + 2] = (*lpix & 0x00FF0000) >> 16;
			lpix++;
		}
		break;
	case VIDEO_PALETTE_RGB24:{
			// uncompressed data. rgb decoder in module simple copy for display.
			memcpy(dst, src, size);
		}
		break;
	default:
		break;
	}
/* rezize16 input rgb24 output rgb565 */
	resize16(pict, dst, owidth, oheight, width, height);
	free(dst);
}

static int isSpcaChip(const char *BridgeName)
{
	int i = -1;
	int size = 0;
	//size = strlen(BridgeName)-1;
	/* Spca506 return more with channel video, cut it */
	//if (size > 10) size = 8;
	/* return Bridge otherwhise -1 */
	for (i = 0; i < MAX_BRIDGE; i++) {
		size = strlen(Blist[i].name);
		if (strncmp(BridgeName, Blist[i].name, size) == 0) {
			printf("Bridge find %s number %d\n", Blist[i].name, i);
			break;
		}
	}

	return i;
}

static int getStreamId(const char *BridgeName)
{
	int i = -1;
	int match = -1;
/* return Stream_id otherwhise -1 */
	if ((match = isSpcaChip(BridgeName)) < 0) {
		printf("Not an Spca5xx Camera !!\n");
		return match;
	}
	switch (match) {
	case BRIDGE_SPCA505:
	case BRIDGE_SPCA506:
		i = YYUV;
		break;
	case BRIDGE_SPCA501:
		i = YUYV;
		break;
	case BRIDGE_SPCA508:
		i = YUVY;
		break;
	case BRIDGE_SPCA504:
	case BRIDGE_SPCA500:
	case BRIDGE_SPCA504B:
	case BRIDGE_SPCA533:
	case BRIDGE_SPCA504C:
	case BRIDGE_SPCA536:
	case BRIDGE_ZR364XX:
	case BRIDGE_ZC3XX:
	case BRIDGE_CX11646:
	case BRIDGE_SN9CXXX:
	case BRIDGE_MR97311:
		i = JPEG;
		break;
	case BRIDGE_ETOMS:
	case BRIDGE_SONIX:
	case BRIDGE_SPCA561:
	case BRIDGE_TV8532:
		i = GBRG;
		break;
	default:
		i = -1;
		printf("Unable to find a StreamId !!\n");
		break;

	}
	return i;
}

static int probeSize(const char *BridgeName, int *width, int *height)
{
	int bridge = -1;
	int i;
	unsigned int intwidth;
	unsigned int intheight;
	unsigned int intformat;
	int match = 0;
/* return 1->ok otherwhise -1 */
	if ((bridge = isSpcaChip(BridgeName)) < 0) {
		printf("Cannot Probe Size !! maybe not an Spca5xx Camera\n");
		return -1;
	}

	for (i = 0; (unsigned int)(GET_EXT_MODES(bridge)[i][0]); i++) {
		intwidth = GET_EXT_MODES(bridge)[i][0];
		intheight = GET_EXT_MODES(bridge)[i][1];
		intformat = (GET_EXT_MODES(bridge)[i][2] & 0xF0) >> 4;
		if ((intwidth == *width) && (intheight == *height)) {
			match = 1;
		} else {
			match = 0;
		}

		printf("intformat:%d\n", intformat);
		printf("Available Resolutions width %d  heigth %d %s %c\n",
		       intwidth, intheight, (intformat) ? "decoded" : "native",
		       (match) ? '*' : ' ');
	}

}

