#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>
#include <linux/types.h>
#include <linux/videodev.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <pthread.h>

#include "video.h"

static pthread_t thread_id;
static unsigned int picnum=0;

static void* save_img(void* args);

int set_vd_size(struct zc3xx_cam *cam, int width, int height)
{
	int vfd;
	struct video_window *win;

	if (cam == NULL)
		return -1;

	vfd = cam->cfd;
	win = &(cam->videowin);

	if (ioctl(vfd, VIDIOCGWIN, win) < 0) {
		PRINT_ERR("failed VIDIOCGWIN");
		return -1;
	}

	win->width = cam->width = width;
	win->height = cam->height = height;
	win->clipcount = 0;

	if (ioctl(vfd, VIDIOCSWIN, win) < 0) {
		PRINT_ERR("failed VIDIOCSWIN");
		return -1;
	}

	cam->img_len = cam->width * cam->height;
	cam->img_buf =
	    (char *)realloc(cam->img_buf, sizeof(char) * cam->img_len);
	//cam->img_buf = (char*)malloc(sizeof(char)*20000);
	if (cam->img_buf == NULL)
		return -1;
	PRINT_DEBUG("set image buffer size %d", cam->img_len);
	return 0;
}

enum {VGA=0, CIF, SCIF};

int change_img_mode (int mode, struct zc3xx_cam *cam, struct pic_struct *pic) 
{ 
	switch(mode) { 
	case VGA: 
	case CIF:
	default:
	break;
	}
	return 0;
}

void grab_release(struct zc3xx_cam *cam) 
{
	if (cam->cfd > 0) 
		close(cam->cfd);
	if (cam->img_buf) 
		free(cam->img_buf);
	if (cam->filename) 
		free(cam->filename);
}

int grab_init(struct zc3xx_cam *cam)
{
	int vfd;
	struct video_channel vch;
	struct video_picture pict;
	struct video_capability vcap;

	if (cam == NULL) {
		return -1;
	}

	cam->cfd = open(cam->devpath, O_RDWR);
	vfd = cam->cfd;
	if (cam->cfd < 0) {
		PRINT_ERR("failed open %s", cam->devpath);
		return -1;
	}

	if (ioctl(vfd, VIDIOCGCAP, &(cam->videocap)) < 0) {
		PRINT_ERR("failed VIDIOCGCAP");
		return -1;
	}
	cam->videocap.name[strlen(cam->videocap.name) - 1] = '\0';
	PRINT_INFO("Found Camera %s\n", cam->videocap.name);
	PRINT_DEBUG("channels: %d\nmaxwidth=%d\tmaxheight=%d\n"
		    "minwidth=%d\tminheight=%d\n", cam->videocap.channels,
		    cam->videocap.maxwidth, cam->videocap.maxheight,
		    cam->videocap.minwidth, cam->videocap.minheight);

	vch.channel = 1;
	if (ioctl(vfd, VIDIOCGCHAN, &vch) < 0) {
		PRINT_ERR("failed VIDIOGCHAN");
		return -1;
	}
	PRINT_INFO("channel:%d name:%s type:%d\n",
		   vch.channel, vch.name, vch.type);
	//cam->videochan->channel, cam->videochan->name, cam->type);

	if (!(vch.type & VIDEO_TYPE_CAMERA)) {
		PRINT_ERR("NOT VID_TYPE_CAPTURE");
		exit(1);
	}

	if (ioctl(vfd, VIDIOCGPICT, &pict) < 0) {
		perror("failed VIDIOCGPICT");
		exit(1);
	}
	pict.palette = VIDEO_PALETTE_JPEG;
	pict.depth = JPEG_FORMAT_DEPTH;
	if (ioctl(vfd, VIDIOCSPICT, &pict) < 0) {
		PRINT_ERR("failed VIDEOCSPICT");
		exit(1);
	}

	if (set_vd_size(cam, cam->width, cam->height) < 0)
		return -1;

	return 0;
}

static void* save_img(void* args) 
{
	int ret;
	FILE* fpic;
	char f_name[512];
	struct pic_struct *pic=(struct pic_struct*)args;

	if (pic == NULL) { 
		PRINT_ERR("pic is NULL");
		exit(1);
	}

	PRINT_INFO("start save_img routin");

	while (1) {
		ret = pthread_mutex_lock(&pic->mutex);
		if (ret < 0) {
			PRINT_ERR("[err][failed pthread_mutex_lock]");
			exit(1);
		}
		while (pic->issave == 0) { 
			PRINT_DEBUG("WAIT");
			ret = pthread_cond_wait(&pic->cond, &pic->mutex);
			PRINT_DEBUG("COND");
			if (ret < 0) {
				PRINT_ERR("[err][failed pthread_cond_wait]");
				exit(1);
			}
		}
		pic->issave = 0;

		if (pic->overwrite)
			sprintf(f_name, "%s.jpg", pic->filename);
		else 
			sprintf(f_name, "%s%03d.jpg", pic->filename, picnum++);
		fpic = fopen(f_name, "wb");
		fwrite(pic->buf, sizeof(char), pic->size, fpic);
		fclose(fpic);

		PRINT_INFO("snapshoot image %s", f_name);
		ret = pthread_mutex_unlock(&pic->mutex);
		if (ret < 0) {
			PRINT_ERR("[err][failed pthread_cond_wait]");
			exit(1);
		}
	}

	return NULL;
}


int start_save_thread(struct pic_struct* pic, char* filename, unsigned int size) 
{
	int ret;

	pic->issave=0;
	pic->filename = strdup(filename);
	pic->buf = (char*)malloc(sizeof(char)*size);
	if (pic->buf == NULL) { 
		return -1;
	}

	ret = pthread_mutex_init(&pic->mutex, NULL);
	if (ret != 0) {
		PRINT_ERR("[ERR][failed pthread_mutex_init]");
		exit(-1);
	}
	ret = pthread_cond_init(&pic->cond, NULL);
	if (ret != 0) {
		PRINT_ERR("[ERR][failed pthread_cond_init]");
		exit(-1);
	}
	ret = pthread_create(&thread_id, NULL, (void *)save_img,
				(void*)pic);
	if (ret != 0) {
		PRINT_ERR("[ERR][capture thread create error");
		exit(-1);
	}

	return 0;
}

void stop_save_thread(struct pic_struct* pic) 
{ 
	if (pthread_cancel(thread_id) < 0) { 
		PRINT_ERR("[ERR][failed pthread_cancel]");
		exit(1);
	}
	print_msg(0, "be join");
	pthread_join(thread_id, NULL);
	pthread_cond_destroy(&(pic->cond));
	pthread_mutex_destroy(&(pic->mutex));
	if (pic->buf) 
		free(pic->buf);
	if (pic->filename) 
		free(pic->filename);
}

int request_save_img(struct pic_struct* pic) {
	int ret;

	ret = pthread_mutex_lock(&pic->mutex);
	if (ret < 0) {
		PRINT_ERR("[ERR][failed pthread_mutex_lock");
		exit(1);
	}
	ret = pthread_cond_signal(&pic->cond);
	if (ret < 0) {
		PRINT_ERR("[ERR][failed pthread_cond_signal");
		exit(1);
	}
	ret = pthread_mutex_unlock(&pic->mutex);
	if (ret < 0) {
		PRINT_ERR("[ERR][failed pthread_mutex_unlock");
		exit(1);
	}
	return 0;
}

